from blackbox_mpc.optimizers.random_search import RandomSearchOptimizer
from blackbox_mpc.optimizers.spsa import SPSAOptimizer
from blackbox_mpc.optimizers.cma_es import CMAESOptimizer
from blackbox_mpc.optimizers.cem import CEMOptimizer
from blackbox_mpc.optimizers.pi2 import PI2Optimizer
from blackbox_mpc.optimizers.pso import PSOOptimizer
from blackbox_mpc.optimizers.optimizer_base import OptimizerBase



import numpy as np
import tensorflow as tf
import logging
from blackbox_mpc.utils.transforms import default_transform_targets, \
    default_inverse_transform_targets


class SystemDynamicsHandler(object):
    def __init__(self, env_action_space, env_observation_space,
                 dynamics_function=None, true_model=False,
                 is_normalized=True,
                 log_dir=None, tf_writer=None,
                 save_model_frequency=1,
                 saved_model_dir=None,
                 transform_targets_func=default_transform_targets,
                 inverse_transform_targets_func=
                 default_inverse_transform_targets):
        """
        This is the system dynamics handler class that is reponsible
        for training the dynamics functions, storing the rollouts as well as
         prepocessing and postprocessing of the MDP elements


        Parameters
        ----------
        dynamics_function: DeterministicDynamicsFunctionBaseClass
            Defines the system dynamics function.
        env_action_space: gym.ActionSpace
            Defines the action space of the gym environment.
        env_observation_space: tf.int32
            Defines the observation space of the gym environment.
        tf_writer: tf.summary
            Defines a tensorflow writer to be used for logging
        log_dir: string
            Defines the log directory to save the normalization statistics in.
        saved_model_dir: string
            Defines the saved model directory where the model is saved in, in
            case of loading the model.
        dynamics_function: DeterministicDynamicsFunctionBase
            Defines the dynamics_function of the nn dynamics function itself
        transform_targets_func: tf_function
            Defines a tf function to transform the next states as targets
            (output of the nn dynamics), by default
            this is the deviation function which is
            (target = next_state - current_state).
        inverse_transform_targets_func: tf_function
            Defines a tf function to inverse transform the targets
            (output of the nn dynamics), by default
            this is the inverse of the deviation function which is
            (next_state = target + current_state).
        save_model_frequency: Int
            Defines how often the model should be saved (defined relative to
            the number of refining iters)
        true_model: bool
            Defines if the dynamics function is a non trainable model or not.
        is_normalized: bool
            Defines if the dynamics function should be trained with
            normalization or not.
        """
        self._is_true_model = true_model
        self._dim_S = tf.constant(env_observation_space.shape[0], dtype=tf.int32)
        self._dim_U = tf.constant(env_action_space.shape[0], dtype=tf.int32)
        self._dynamics_function = dynamics_function
        self._save_model_frequency = save_model_frequency
        self._log_dir = log_dir
        self._tf_writer = tf_writer
        self._saved_model_dir = saved_model_dir
        self._is_normalized = is_normalized
        self._transform_targets = transform_targets_func
        self._inverse_transform_targets = inverse_transform_targets_func
        self._model_training_in = np.array([], dtype=np.float32).reshape(0, self._dim_U + self._dim_S)
        self._model_validation_in = np.array([], dtype=np.float32).reshape(0, self._dim_U + self._dim_S)
        self._model_training_out = np.array([], dtype=np.float32).reshape(0, self._dim_S)
        self._model_validation_out = np.array([], dtype=np.float32).reshape(0, self._dim_S)
        self._training_iter = 0
        self._refining_model_iter = 0
        self._first_time = True
        if self._saved_model_dir is not None:
            logging.info("Loading the saved model now....")
            self._dynamics_function = tf.saved_model.load(self._saved_model_dir)
            #load the stats and set first time to false
            self._first_time = False
            if self._is_normalized:
                self._mean_states = np.load(self._saved_model_dir +
                                            '/mean_states.npy')
                self._std_states = np.load(self._saved_model_dir +
                                           '/std_states.npy')
                self._mean_actions = np.load(self._saved_model_dir +
                                             '/mean_actions.npy')
                self._std_actions = np.load(self._saved_model_dir +
                                            '/std_actions.npy')
                self._mean_targets = np.load(self._saved_model_dir +
                                             '/mean_targets.npy')
                self._std_targets = np.load(self._saved_model_dir +
                                            '/std_targets.npy')

    @tf.function
    def process_input(self, states, actions):
        """
        This is the process_input function, which takes in the states and
        actions and preprocesses them for the dynamics
        function, (normalization..etc)

        Parameters
        ---------
        states: tf.float32
            The current states has a shape of (Batch Xdim_S)
        actions: tf.float32
            The current actions has a shape of (Batch Xdim_U)

        Returns
        -------
        result: tf.float32
            concatenated states and actions after preprocessing them.
        """
        if self._is_true_model:
            return tf.concat([states, actions], axis=-1)
        else:
            if self._is_normalized:
                new_states = (states - self._mean_states) / (
                            self._std_states + 1e-7)
                new_actions = (actions - self._mean_actions) / (
                            self._std_actions + 1e-7)
                return tf.concat([new_states, new_actions], axis=-1)
            else:
                return tf.concat([states, actions], axis=-1)

    @tf.function
    def process_output(self, inputs_states, raw_output):
        """
        This is the process_state_output function, which takes in the
        previous states predicted target/delta and processes
        them to get the predicted absolute next state.

        Parameters
        ---------
        inputs_states: tf.float32
            The previous states has a shape of (Batch Xdim_S)
        raw_output: tf.float32
            The predicted normalized delta as received from the dynamics function
            has a shape of (Batch Xdim_U).

        Returns
        -------
        result: tf.float32
            absolute predicted next state.
        """
        if self._is_true_model:
            new_states_transformed = \
                self._inverse_transform_targets(inputs_states, raw_output)
            return new_states_transformed
        else:
            if self._is_normalized:
                new_states_deviation = self._mean_targets + (
                        raw_output * (
                                self._std_targets + 1e-7))
            else:
                new_states_deviation = raw_output
            new_states = self._inverse_transform_targets(inputs_states,
                                                         new_states_deviation)
            return new_states

    def train(self, observations_trajectories, actions_trajectories,
              rewards_trajectories, validation_split=0.2,
              batch_size=128, learning_rate=1e-3, epochs=30,
              nn_optimizer=tf.keras.optimizers.Adam):
        """
        This is the train function, which takes in the data of the MDP to
        train the dynamics model on it.

        Parameters
        ---------
        observations_trajectories: [np.float32]
            A list of observations of each of the episodes for the n agents.
        actions_trajectories: [np.float32]
            A list of actions of each of the episodes for the n agents.
        rewards_trajectories: [np.float32]
            A list of rewards of each of the episodes for the n agents.
        learning_rate: float
            Learning rate to be used in training the dynamics function.
        epochs: Int
            Number of epochs to be used in training the dynamics function
            everytime train is called.
        validation_split: float32
            Defines the validation split to be used of the rollouts collected.
        batch_size: int
            Defines the batch size to be used for training the model.
        nn_optimizer: tf.keras.optimizers
            Defines the optimizer to use with the neural network.
        """
        self._append_to_training_dataset(observations_trajectories, actions_trajectories,
                                         rewards_trajectories, validation_split=validation_split)
        if self._first_time: #TODO: for some reason this is needed
            if self._is_normalized:
                self._recompute_normalization()
                self._first_time = False
            else:
                self._first_time = False
        model_training_in, model_training_out = self._normalize_data(self._model_training_in, self._model_training_out)
        train_dataset = tf.data.Dataset.from_tensor_slices((model_training_in,
                                                            model_training_out)).\
            shuffle(model_training_in.shape[0]).batch(batch_size, drop_remainder=True)
        model_validation_in, model_validation_out = self._normalize_data(self._model_validation_in,
                                                                         self._model_validation_out)
        validation_dataset = tf.data.Dataset.from_tensor_slices((model_validation_in,
                                                                 model_validation_out)).\
            batch(batch_size, drop_remainder=True)
        logging.info("Started the system training")
        self._training_algorithm(train_dataset, validation_dataset,
                                 learning_rate=learning_rate, epochs=epochs,
                                 nn_optimizer=nn_optimizer)
        self._training_iter += 1
        if self._training_iter % self._save_model_frequency == 0 and \
                self._log_dir is not None:
            logging.info("Saving the model now....")
            call = self._dynamics_function.__call__.get_concrete_function(
                tf.TensorSpec([None, self._dim_U + self._dim_S], tf.float32),
                tf.TensorSpec([], tf.bool))
            tf.saved_model.save(self._dynamics_function, self._log_dir +
                                '/saved_model_' +
                                str(self._refining_model_iter) + '/',
                                signatures=call)
            #save the means and std devs
            np.save(self._log_dir + '/saved_model_' +
                                str(self._refining_model_iter) + '/mean_states',
                    self._mean_states)
            np.save(self._log_dir + '/saved_model_' +
                                str(self._refining_model_iter) + '/std_states',
                    self._std_states)
            np.save(self._log_dir + '/saved_model_' +
                                str(self._refining_model_iter) + '/mean_actions',
                    self._mean_actions)
            np.save(self._log_dir + '/saved_model_' +
                                str(self._refining_model_iter) + '/std_actions',
                    self._std_actions)
            np.save(self._log_dir + '/saved_model_' +
                                str(self._refining_model_iter) + '/mean_targets',
                    self._mean_targets)
            np.save(self._log_dir + '/saved_model_' +
                                str(self._refining_model_iter) + '/std_targets',
                    self._std_targets)
        logging.info("Ended the system training")
        return

    def _training_algorithm(self, train_dataset, validation_dataset,
                            learning_rate=1e-3, epochs=30,
                            nn_optimizer=tf.keras.optimizers.Adam):
        """
        This is the train_model function for the dynamics trainer base class.


        Parameters
        ---------
        train_dataset: tf.data.Dataset
            training dataset that is composed of (input, expected_output)
            after processing.
        validation_dataset: tf.data.Dataset
            validation dataset that is composed of (input, expected_output)
            after processing.
        """
        optimizer = nn_optimizer(learning_rate=learning_rate)
        training_loss = np.zeros(epochs)
        validation_loss = np.zeros(epochs)
        for i in range(0, epochs):  # run through epochs
            average_loss = 0
            number_of_batches = 0
            for x, y in train_dataset:
                number_of_batches += 1
                with tf.GradientTape() as tape:
                    prediction = \
                        self._dynamics_function(x,
                                                train=tf.constant(
                                                    True,
                                                    dtype=tf.bool))
                    loss = \
                        self._dynamics_function.get_loss(expected_output=y,
                                                         predictions=prediction)
                    average_loss += loss
                trainable_variables = tape.watched_variables()
                gradients = tape.gradient(loss, trainable_variables)
                optimizer.apply_gradients(zip(gradients, trainable_variables))

            training_loss[i] = average_loss / number_of_batches
            average_loss = 0
            number_of_batches = 0
            for x, y in validation_dataset:
                number_of_batches += 1
                prediction = self._dynamics_function(x,
                                                     train=tf.constant(False, dtype=tf.bool))
                loss = self._dynamics_function.get_validation_loss(
                    expected_output=y, predictions=prediction)
                average_loss += loss
            validation_loss[i] = average_loss / number_of_batches
            if self._tf_writer is not None:
                with self._tf_writer.as_default():
                    tf.summary.scalar('system_model_val/loss', validation_loss[i],
                                      step=(self._refining_model_iter * epochs) + i)
        self._refining_model_iter += 1

    def _append_to_training_dataset(self, observations_trajectories,
                                    actions_trajectories,
                                    rewards_trajectories,
                                    validation_split=0.2):
        actions_trajectories = np.array(actions_trajectories)
        num_of_agents = actions_trajectories.shape[2]
        new_data_in, new_data_targs = [], []
        for obs, acs in zip(observations_trajectories, actions_trajectories):
            #switch number of runner with traj length
            for agent in range(num_of_agents):
                states = obs[:-1, agent]
                actions = acs[:, agent]
                new_data_in.append(np.concatenate([states, actions], axis=-1))
                next_states = obs[1:, agent]
                new_data_targs.append(self._transform_targets(states,
                                                              next_states).numpy())
        #now split data into val and train
        new_data_in = np.array(new_data_in, dtype=np.float32).reshape(-1, self._dim_U + self._dim_S)
        new_data_targs = np.array(new_data_targs, dtype=np.float32).reshape(-1, self._dim_S)
        training_indicies = np.random.choice([False, True],
                                             size=new_data_in.shape[0],
                                             p=[validation_split, 1.0 - validation_split])
        new_train_in = new_data_in[training_indicies]
        new_train_targs = new_data_targs[training_indicies]
        new_validation_in = new_data_in[~training_indicies]
        new_validation_targs = new_data_targs[~training_indicies]
        N_train = new_train_in.shape[0]
        N_val = new_validation_in.shape[0]
        self._model_training_in = np.concatenate([self._model_training_in, new_train_in], axis=0)
        self._model_training_out = np.concatenate([self._model_training_out, new_train_targs], axis=0)
        self._model_validation_in = np.concatenate([self._model_validation_in, new_validation_in], axis=0)
        self._model_validation_out = np.concatenate([self._model_validation_out, new_validation_targs], axis=0)
        return

    def _normalize_data(self, data_in, data_out):
        states_normalized = (data_in[:, :self._dim_S] - self._mean_states) / (self._std_states + 1e-7)
        actions_normalized = (data_in[:, self._dim_S:] - self._mean_actions) / (self._std_actions + 1e-7)
        targets_normalized = (data_out - self._mean_targets) / (self._std_targets + 1e-7)
        return np.concatenate([states_normalized, actions_normalized], axis=1), targets_normalized

    def _recompute_normalization(self):
        self._mean_states = np.mean(self._model_training_in[:, :self._dim_S], axis=0)
        self._std_states = np.std(self._model_training_in[:, :self._dim_S], axis=0)

        self._mean_actions = np.mean(self._model_training_in[:, self._dim_S:], axis=0)
        self._std_actions = np.std(self._model_training_in[:, self._dim_S:], axis=0)

        self._mean_targets = np.mean(self._model_training_out, axis=0)
        self._std_targets = np.std(self._model_training_out, axis=0)
        return

    def get_dynamics_function(self):
        """
        returns the dynamics function used by the system handler.

        :return:
        """
        return self._dynamics_function



class ModelBasedBasePolicy(object):
    def __init__(self, trajectory_evaluator):
        """
            This is the model based policy base class for controlling the agent

        Parameters
        ---------
        trajectory_evaluator: EvaluatorBase
            Defines the trajectory evaluator to be used in the optimizer to
            evaluate trajectories.
        """
        self._trajectory_evaluator = trajectory_evaluator
        return

    def act(self,  observations, t, exploration_noise=False):
        """
        This is the act function for the model based policy base class,
        which should be called to provide the action
        to be executed at the current time step.


        Parameters
        ---------
        observations: tf.float32
            Defines the current observations received from the environment.
        t: tf.float32
            Defines the current timestep.
        exploration_noise: bool
            Defines if exploration noise should be added to the action to be executed.


        Returns
        -------
        action: tf.float32
            The action to be executed for each of the runner (dims = runner X dim_U)
        next_observations: tf.float32
            The next observations predicted using the dynamics function learned so far.
        rewards_of_next_state: tf.float32
            The predicted reward if the action was executed using the predicted observations.
        """
        raise Exception("act function is not implemented yet")

    def reset(self):
        """
        This is the reset function for the model based policy base class,
        which should be called at the beginning of the episode.
        """
        raise Exception("reset function is not implemented yet")



import tensorflow as tf
import numpy as np
from blackbox_mpc.optimizers.optimizer_base import OptimizerBase


class RandomSearchOptimizer(OptimizerBase):
    def __init__(self, env_action_space, env_observation_space,
                 planning_horizon=50, population_size=1024, num_agents=5):
        """
        This class is responsible for performing random shooting and choosing the best
        possible predicted trajectory and returning the first action of this trajectory.


        Parameters
        ---------
        env_action_space: gym.ActionSpace
            Defines the action space of the gym environment.
        env_observation_space: gym.ObservationSpace
            Defines the observation space of the gym environment.
        planning_horizon: Int
            Defines the planning horizon for the optimizer (how many steps to lookahead and optimize for).
        population_size: tf.int32
            Defines the population size of the particles evaluated at each iteration.
        num_agents: tf.int32
            Defines the number of runner running in parallel
        """
        super(RandomSearchOptimizer, self).__init__(name=None,
                                                    planning_horizon=planning_horizon,
                                                    max_iterations=None,
                                                    num_agents=num_agents,
                                                    env_action_space=env_action_space,
                                                    env_observation_space=
                                                    env_observation_space)
        self._solution_dim = [self._num_agents, self._planning_horizon, self._dim_U]
        self._population_size = population_size
        return

    @tf.function
    def _optimize(self, current_state, time_step):
        samples = tf.random.uniform([self._population_size, *self._solution_dim], self._action_lower_bound_horizon,
                                    self._action_upper_bound_horizon, dtype=tf.float32)
        rewards = self._trajectory_evaluator(current_state, samples, time_step)
        best_particle_index = tf.cast(tf.math.argmax(rewards), dtype=tf.int32)
        samples = tf.transpose(samples, [1, 0, 2, 3])
        best_particle_index = best_particle_index + tf.range(0, samples.shape[0], dtype=tf.int32)*samples.shape[1]
        samples = tf.reshape(samples, [-1, *samples.shape[2:]])
        resulting_action = tf.gather(samples, best_particle_index)[:, 0]
        return resulting_action

    def reset(self):
        """
          This method resets the optimizer to its default state at the beginning of the trajectory/episode.
          """
        return



import numpy as np
import multiprocessing as mp
import gym


class EnvWorker(mp.Process):
    """This is the Env Worker class which is responsible for running one agent in one process"""
    def __init__(self, remote, env_fn, queue, lock):
        """
        This is the initializer function for the Env Worker class.

        Parameters
        ---------
        remote: multiprocessing.Pipe()
            The remote process responsible for sending and receiving to/from the parent process.
        env_fn: py_func
            A python function that would return a new environment when called
        queue: multiprocessing.Queue()
            A multiprocessing queue to be used in the cosumption of shared resources.
        lock: multiprocessing.Lock()
             A lock to be used in locking shared resources between agents.
        """
        super(EnvWorker, self).__init__()
        self.remote = remote
        self.env = env_fn()
        self.queue = queue
        self.lock = lock
        self.done = False

    def empty_step(self):
        """
        This is the empty step function for the Env Worker class, a step without an action basically.

        Returns
        -------
        observation: np.float32
            The observations received with a zeros in the action as a step in the environment.
        reward: np.float32
            The rewards received with a zeros in the action as a step in the environment.
        done: bool
            A boolean if true then the worker is terminated, if not then the worker is not terminated.
        info: empty dict
            An empty dict to keep the consistency with the other step.

        """
        observation = np.zeros(self.env.observation_space.shape,
                               dtype=np.float32)
        reward, done = 0.0, True
        return observation, reward, done, {}

    def try_reset(self):
        """
        This is the try reset function for the Env Worker class, which resets the environment of the worker.

        Returns
        -------
        observation: np.float32
            The observations received after a reset in the environment.
        """
        observation = self.env.reset()
        return observation

    def run(self):
        """
        This is the run function of the worker which sends and receives data.

        - If command is step then

        Returns
        -------
        observation: np.float32
            The observations after executing the actions received.
        reward: np.float32
            The rewards after executing the actions received.
        done: bool
            A boolean if true then the worker is terminated, if not then the worker is not terminated.
        info: dict
            A dict that returns extra information of the step that was just executed in the environment.



        - If command is reset then

        Returns
        -------
        observation: np.float32
            The observations after a reset in the environment.


        - If command is get_spaces then

        Returns
        -------
        observation_space: ObservationSpaceObject
            The observation space object which has a low and a high as well as shape.
        action_space: ActionSpaceObject
            The action space object which has a low and a high as well as shape.



        - If command is render then

        Returns
        -------
        result: np.float32
            returns the RGB array of rendering the environment at this current timestep.

        """
        while True:
            command, data = self.remote.recv()
            if command == 'step':
                observation, reward, done, info = self.env.step(data)
                self.remote.send((observation, reward, done, info))
            elif command == 'reset':
                observation = self.try_reset()#get new id for the worker here
                self.remote.send((observation))
            elif command == 'close':
                self.remote.close()
                break
            elif command == 'get_spaces':
                self.remote.send((self.env.observation_space,
                                 self.env.action_space))
            elif command == 'render':
                self.remote.send(self.env.render(mode='rgb_array'))
            else:
                raise NotImplementedError()


class SubprocVecEnv(gym.Env):
    """This is the SubprocVecEnv class which is responsible for running n agents in n processes"""
    metadata = {'render.modes': ['rgb_array']}

    def __init__(self, env_factory, queue):
        """
        This is the initializer function for the SubprocVecEnv class.

        Parameters
        ---------
        env_factory: [py_func]
            A list of py_funcs where each py_func returns a new environment for a new process.
        queue: multiprocessing.Queue()
            A multiprocessing queue to be used in the cosumption of shared resources.
        """
        self.lock = mp.Lock()
        self.remotes, self.work_remotes = zip(*[mp.Pipe() for _ in env_factory])
        self.workers = [EnvWorker(remote, env_fn, queue, self.lock)
            for (remote, env_fn) in zip(self.work_remotes, env_factory)]
        for worker in self.workers:
            worker.daemon = True
            worker.start()
        for remote in self.work_remotes:
            remote.close()
        self.waiting = False
        self.closed = False

        self.remotes[0].send(('get_spaces', None))
        observation_space, action_space = self.remotes[0].recv()
        self.observation_space = observation_space
        self.action_space = action_space

    def step(self, actions):
        """
        This is the step function for the SubprocVecEnv class.

        Parameters
        ---------
        actions: np.float32
            Actions to be executed to each of the environment_utils/agents.


        Returns
        -------
        observation: np.float32
            The observations after executing the actions received.
        reward: np.float32
            The rewards after executing the actions received.
        done: [bool]
            A list of boolean if true then the agent is terminated, if not then the agent is not terminated.
        info: dict
            A dict that returns extra information of the step that was just executed in the parallel environment_utils.
        """
        self._step_async(actions)
        return self._step_wait()

    def _step_async(self, actions):
        for remote, action in zip(self.remotes, actions):
            remote.send(('step', action))
        self.waiting = True

    def _step_wait(self):
        results = [remote.recv() for remote in self.remotes]
        self.waiting = False
        observations, rewards, dones, infos = zip(*results)
        return np.stack(observations), np.stack(rewards), np.stack(dones), infos

    def reset(self):
        """
        This is the reset function for the SubprocVecEnv class.

        Returns
        -------
        observation: np.float32
            The observations after resetting each of the environment_utils.
        """
        for remote in self.remotes:
            remote.send(('reset', None))
        results = [remote.recv() for remote in self.remotes]
        return np.stack(results)

    def close(self):
        """
        This is the close function for the SubprocVecEnv class.
        """
        if self.closed:
            return
        if self.waiting:
            for remote in self.remotes:
                remote.recv()
        for remote in self.remotes:
            remote.send(('close', None))
        for worker in self.workers:
            worker.join()
        self.closed = True

    def render(self, mode='human'):
        """
        This is the render function for the SubprocVecEnv class.

        Returns
        -------
        image: np.float32
            The RGB array of the current view of the environment_utils.
        """
        for remote in self.remotes:              
            remote.send(('render', None))
        results = [remote.recv() for remote in self.remotes]
        return results



import tensorflow as tf
import numpy as np
from blackbox_mpc.optimizers.optimizer_base import OptimizerBase


class PSOOptimizer(OptimizerBase):
    def __init__(self, env_action_space, env_observation_space,
                 planning_horizon=50, max_iterations=5, population_size=500,
                 num_agents=5, c1=tf.constant(0.3, dtype=tf.float32),
                 c2=tf.constant(0.5, dtype=tf.float32), w=tf.constant(0.2, dtype=tf.float32),
                 initial_velocity_fraction=tf.constant(0.01, dtype=tf.float32)):
        """
        This class defines the particle swarm optimizer.
        (https://www.cs.tufts.edu/comp/150GA/homeworks/hw3/_reading6%201995%20particle%20swarming.pdf)


        Parameters
        ---------
        env_action_space: gym.ActionSpace
            Defines the action space of the gym environment.
        env_observation_space: gym.ObservationSpace
            Defines the observation space of the gym environment.
        planning_horizon: Int
            Defines the planning horizon for the optimizer (how many steps to lookahead and optimize for).
        max_iterations: tf.int32
            Defines the maximimum iterations for the CMAES optimizer to refine its guess for the optimal solution.
        population_size: tf.int32
            Defines the population size of the particles evaluated at each iteration.
        num_agents: tf.int32
            Defines the number of runner running in parallel
        c1: tf.float32
            Defines the fraction of the local best known position direction.
        c2: tf.float32
            Defines the fraction of the global best known position direction.
        w: tf.float32
            Defines the fraction of the current velocity to use.
        initial_velocity_fraction: tf.float32
           Defines the initial velocity fraction out of the action space.
        """
        super(PSOOptimizer, self).__init__(name=None,
                                           planning_horizon=planning_horizon,
                                           max_iterations=max_iterations,
                                           num_agents=num_agents,
                                           env_action_space=env_action_space,
                                           env_observation_space=
                                           env_observation_space)
        self._solution_dim = [self._num_agents, tf.constant(self._planning_horizon, dtype=tf.int32), self._dim_U]
        self._solution_size = tf.reduce_prod(self._solution_dim)
        self._population_size = population_size
        self._particle_positions = tf.Variable(tf.zeros([self._population_size, *self._solution_dim], dtype=tf.float32))
        self._particle_velocities = tf.Variable(tf.zeros([self._population_size, *self._solution_dim], dtype=tf.float32))
        self._particle_best_known_position = tf.Variable(tf.zeros([self._population_size, *self._solution_dim],
                                                                  dtype=tf.float32))
        self._particle_best_known_reward = tf.Variable(tf.zeros([self._population_size, self._num_agents],
                                                                dtype=tf.float32))

        #global
        self._global_best_known_position = tf.Variable(tf.zeros([*self._solution_dim], dtype=tf.float32))
        self._global_best_known_reward = tf.Variable(tf.zeros([self._num_agents], dtype=tf.float32))
        solution_variance_values = np.tile(np.square(self._action_lower_bound - self._action_upper_bound) / 16,
                                           [self._planning_horizon * self._num_agents, 1])
        solution_variance_values = solution_variance_values.reshape([self._num_agents, self._planning_horizon, -1])
        self._solution_variance = tf.constant(solution_variance_values, dtype=tf.float32)
        self._c1 = c1
        self._c2 = c2
        self._w = w
        self._initial_velocity_fraction = initial_velocity_fraction
        self._solution = tf.Variable(tf.zeros([self._num_agents, self._dim_U], dtype=tf.float32))

    @tf.function
    def _optimize(self, current_state, time_step):
        def continue_condition(t, position):
            result = tf.less(t, self._max_iterations)
            return result

        def iterate(t, position):
            #evaluate each of the particles
            # Evaluate and sort solutions
            feasible_particle_positions = tf.clip_by_value(self._particle_positions, self._action_lower_bound_horizon,
                                                           self._action_upper_bound_horizon)
            penalty = tf.norm(tf.reshape(self._particle_positions - feasible_particle_positions, [self._population_size, self._num_agents, -1]),
                              axis=2) ** 2
            self._particle_positions.assign(feasible_particle_positions)

            rewards = self._trajectory_evaluator(current_state, self._particle_positions, time_step) - penalty
            #set the best local known position
            condition = tf.less(self._particle_best_known_reward, rewards)

            new_particle_best_known_position = tf.where(tf.expand_dims(tf.expand_dims(condition, -1), -1), self._particle_positions,
                                                        self._particle_best_known_position)
            self._particle_best_known_position.assign(new_particle_best_known_position)
            new_particle_best_known_reward = tf.where(condition, rewards,
                                                      self._particle_best_known_reward)
            self._particle_best_known_reward.assign(new_particle_best_known_reward)
            #get the global best now

            global_best_known_position_index = tf.math.argmax(self._particle_best_known_reward)
            samples = tf.transpose(self._particle_best_known_position, [1, 0, 2, 3])
            global_best_known_position_index = tf.cast(global_best_known_position_index, dtype=tf.int32) + tf.range(0, samples.shape[0], dtype=tf.int32) * samples.shape[1]
            samples = tf.reshape(samples, [-1, *samples.shape[2:]])
            self._global_best_known_position.assign(tf.gather(samples, global_best_known_position_index))
            samples = tf.reshape(self._particle_best_known_reward, [-1])
            self._global_best_known_reward.assign(tf.gather(samples, global_best_known_position_index))


            #calculate the velocity now
            adapted_particle_velocities = (self._particle_velocities * self._w) + \
                                          (self._particle_best_known_position - self._particle_positions) * self._c1 * tf.random.normal(shape=[], dtype=tf.float32) + \
                                          (self._global_best_known_position - self._particle_positions) * self._c2 * tf.random.normal(shape=[], dtype=tf.float32)
            self._particle_velocities.assign(adapted_particle_velocities)
            self._particle_positions.assign(self._particle_positions + self._particle_velocities)
            return t + tf.constant(1, dtype=tf.int32), self._global_best_known_position
        _ = tf.while_loop(cond=continue_condition, body=iterate, loop_vars=[tf.constant(0, dtype=tf.int32), self._global_best_known_position])
        self._solution.assign(self._global_best_known_position[:, 0, :])
        # update the particles position for the next iteration
        lower_bound_dist = self._global_best_known_position - self._action_lower_bound_horizon
        upper_bound_dist = self._action_upper_bound_horizon - self._global_best_known_position
        constrained_variance = tf.minimum(tf.minimum(tf.square(lower_bound_dist / tf.constant(2, dtype=tf.float32)),
                                                     tf.square(upper_bound_dist / tf.constant(2, dtype=tf.float32))),
                                          self._solution_variance)
        samples_positions = tf.random.truncated_normal([self._population_size,
                                                        *self._solution_dim],
                                                       tf.concat([self._global_best_known_position[:, 1:],
                                                                  tf.expand_dims(self._global_best_known_position[:, -1],
                                                                                 1)], 1),
                                                       tf.sqrt(constrained_variance),
                                                       dtype=tf.float32)
        action_space = self._action_upper_bound_horizon - self._action_lower_bound_horizon
        initial_velocity = self._initial_velocity_fraction * action_space
        samples_velocities = tf.random.uniform([self._population_size, *self._solution_dim], -initial_velocity,
                                               initial_velocity, dtype=tf.float32)
        self._particle_positions.assign(samples_positions)
        self._particle_velocities.assign(samples_velocities)
        self._particle_best_known_position.assign(samples_positions)
        self._particle_best_known_reward.assign(tf.fill([self._population_size, self._num_agents],
                                                        tf.constant(-np.inf, dtype=tf.float32)))
        self._global_best_known_reward.assign(tf.fill([self._num_agents],
                                                      tf.constant(-np.inf, dtype=tf.float32)))
        #end update particles
        resulting_action = self._solution
        return resulting_action

    def reset(self):
        """
         This method resets the optimizer to its default state at the beginning of the trajectory/episode.
         """
        samples_positions = tf.random.uniform([self._population_size, *self._solution_dim], self._action_lower_bound_horizon,
                                              self._action_upper_bound_horizon, dtype=tf.float32)
        action_space = self._action_upper_bound_horizon - self._action_lower_bound_horizon
        initial_velocity = self._initial_velocity_fraction * action_space
        samples_velocities = tf.random.uniform([self._population_size, *self._solution_dim], -initial_velocity,
                                               initial_velocity, dtype=tf.float32)
        self._particle_positions.assign(samples_positions)
        self._particle_velocities.assign(samples_velocities)
        self._particle_best_known_position.assign(samples_positions)
        self._particle_best_known_reward.assign(tf.fill([self._population_size, self._num_agents],
                                                        tf.constant(-np.inf, dtype=tf.float32)))
        self._global_best_known_reward.assign(tf.fill([self._num_agents],
                                                      tf.constant(-np.inf, dtype=tf.float32)))
        return





import tensorflow as tf
import numpy as np


class OptimizerBase(tf.Module):
    def __init__(self, name, planning_horizon, max_iterations, num_agents,
                 env_action_space, env_observation_space):
        """
        This is the base class of the optimizers


        Parameters
        ---------
        name: String
            Defines the name of the block of the optimizer.
        planning_horizon: Int
            Defines the planning horizon for the optimizer (how many steps to lookahead and optimize for).
        max_iterations: tf.int32
            Defines the maximimum iterations for the CEM optimizer to refine its guess for the optimal solution.
        num_agents: tf.int32
            Defines the number of runner running in parallel
        env_action_space: gym.ActionSpace
            Defines the action space of the gym environment.
        env_observation_space: tf.int32
            Defines the observation space of the gym environment.
        """
        super(OptimizerBase, self).__init__(name=name)
        self._planning_horizon = planning_horizon
        self._env_action_space = env_action_space
        self._env_observation_space = env_observation_space
        self._dim_U = tf.constant(env_action_space.shape[0], dtype=tf.int32)
        self._dim_S = tf.constant(env_observation_space.shape[0], dtype=tf.int32)
        self._action_upper_bound = tf.constant(env_action_space.high,
                                               dtype=tf.float32)
        self._action_lower_bound = tf.constant(env_action_space.low,
                                               dtype=tf.float32)
        self._action_upper_bound_horizon = tf.tile(
            np.expand_dims(self._action_upper_bound, 0),
            [self._planning_horizon, 1])
        self._action_lower_bound_horizon = tf.tile(
            np.expand_dims(self._action_lower_bound, 0),
            [self._planning_horizon, 1])
        self._num_agents = num_agents
        self._max_iterations = max_iterations
        self._trajectory_evaluator = None
        self._exploration_variance = (np.square(self._action_lower_bound -
                                                self._action_upper_bound) /
                                      16) * 0.05
        self._exploration_mean = (self._action_upper_bound +
                                  self._action_lower_bound) / 2

    def _optimize(self, current_state, time_step):
        raise Exception("__call__ function is not implemented yet")

    @tf.function
    def __call__(self, current_state, time_step, add_exploration_noise):
        """
       This is the call function for the Base Optimizer Class.
       It is used to calculate the optimal solution for action at the current timestep given the current state.

       Parameters
       ---------
       current_state: tf.float32
           Defines the current state of the system, (dims=num_of_agents X dim_S)
       time_step: tf.float32
           Defines the current timestep of the episode.
       exploration_noise: tf.bool
           Define if the optimal action should have some noise added to it before returning it.


       Returns
       -------
       resulting_action: tf.float32
            The optimal solution for the first action to be applied in the current time step.
       next_state: tf.float32
            The next state predicted using the dynamics model in the trajectory evaluator.
       rewards_of_next_state: tf.float32
            The predicted reward achieved after applying the action given by the optimizer.
       """
        resulting_action = \
            self._optimize(current_state, time_step)
        if add_exploration_noise:
            noise = tf.random.truncated_normal([self._num_agents, self._dim_U],
                                               self._exploration_mean,
                                               tf.sqrt(self._exploration_variance),
                                               dtype=tf.float32)
            resulting_action = resulting_action + noise
            resulting_action = tf.clip_by_value(resulting_action,
                                                self._action_lower_bound,
                                                self._action_upper_bound)
        next_state = self._trajectory_evaluator.predict_next_state(
            current_state, resulting_action)
        rewards_of_next_state = self._trajectory_evaluator.\
            evaluate_next_reward(current_state, next_state, resulting_action)
        return resulting_action, next_state, rewards_of_next_state

    def reset(self):
        """
          This method resets the optimizer to its default state at the
          beginning of the trajectory/episode.
          """

        raise Exception("reset function is not implemented yet")

    def set_trajectory_evaluator(self, trajectory_evaluator):
        """
        Sets the trajectory evaluator to be used by the optimizer.

        :param trajectory_evaluator: (EvaluatorBaseClass) Defines the
                trajectory evaluator to be used to evaluate the reward of a
                sequence of actions.
        :return:
        """
        self._trajectory_evaluator = trajectory_evaluator
        return



from blackbox_mpc.policies.mpc_policy import \
    MPCPolicy
from blackbox_mpc.utils.dynamics_learning import learn_dynamics_from_policy
import logging
logging.getLogger().setLevel(logging.INFO)
import tensorflow as tf
from blackbox_mpc.dynamics_handlers.system_dynamics_handler import \
    SystemDynamicsHandler


def learn_dynamics_iteratively_w_mpc(env,
                                     number_of_initial_rollouts,
                                     number_of_rollouts_for_refinement,
                                     number_of_refinement_steps,
                                     task_horizon,
                                     env_action_space=None,
                                     env_observation_space=None,
                                     initial_policy=None,
                                     refinement_policy=None,
                                     planning_horizon=None,
                                     reward_function=None,
                                     is_normalized=True,
                                     optimizer_name='CEM',
                                     optimizer=None,
                                     num_agents=None,
                                     nn_optimizer=tf.keras.optimizers.Adam,
                                     dynamics_function=None,
                                     system_dynamics_handler=None,
                                     log_dir=None,
                                     tf_writer=None,
                                     save_model_frequency=1,
                                     saved_model_dir=None,
                                     exploration_noise=False,
                                     epochs=30, learning_rate=1e-3,
                                     validation_split=0.2, batch_size=128,
                                     start_episode=0,
                                     **optimizer_args):
    """
    This is the learn dynamics function iteratively using mpc policy
    for the runner class which samples n rollouts using an initial policy and then
    uses these rollouts to learn a dynamics function for the system which is then used to _sample further rollouts
    to refine the dynamics function.


    Parameters
    ---------
    env: parallelgymEnv
        a wrapped gym environment using blackbox.environment_utils.EnvironmentWrapper funcs
    env_action_space: gym.ActionSpace
            Defines the action space of the gym environment.
    env_observation_space: gym.ObservationSpace
        Defines the observation space of the gym environment.
    num_agents: tf.int32
            Defines the number of runner running in parallel
    dynamics_function: DeterministicDynamicsFunctionBaseClass
        Defines the system dynamics function.
    system_dynamics_handler: SystemDynamicsHandler
            The system_dynamics_handler is a handler of the state, actions and
            targets processing funcs as well.
    number_of_initial_rollouts: Int
        Number of initial rollouts/ episodes to perform for each of the agents in the vectorized environment.
    number_of_rollouts_for_refinement: Int
        Number of refinement rollouts/ episodes to perform for each of the agents in the vectorized environment.
    number_of_refinement_steps: Int
        Number of refinemnet steps train, collect, train..etc to run for.
    task_horizon: Int
        The task horizon/ episode length.
    initial_policy: ModelBasedBasePolicy or ModelFreeBasePolicy
        The policy to be used in collecting the initial episodes from the different agents.
    refinement_policy: ModelBasedBasePolicy
        The policy to be used in collecting the followup episodes to refine the policy.
    exploration_noise: bool
        If noise should be added to the actions to help in exploration.
    learning_rate: float
        Learning rate to be used in training the dynamics function.
    epochs: Int
        Number of epochs to be used in training the dynamics function everytime train is called.
    validation_split: float32
        Defines the validation split to be used of the rollouts collected.
    batch_size: int
        Defines the batch size to be used for training the model.
    nn_optimizer: tf.keras.optimizers
        Defines the optimizer to use with the neural network.
    is_normalized: bool
        Defines if the dynamics function should be trained with normalization or not.
    reward_function: tf_function
            Defines the reward function with the prototype: tf_func_name(current_state, current_actions, next_state),
            where current_state is BatchXdim_S, next_state is BatchXdim_S and  current_actions is BatchXdim_U.
    planning_horizon: tf.int32
        Defines the planning horizon for the optimizer (how many steps to lookahead and optimize for).
    optimizer: OptimizerBaseClass
        Optimizer to be used that optimizes for the best action sequence and returns the first action.
    optimizer_name: str
        optimizer name between in ['CEM', 'CMA-ES', 'PI2', 'RandomSearch', 'PSO', 'SPSA'].
    saved_model_dir: string
            Defines the saved model directory where the model is saved in, in case of loading the model.
    save_model_frequency: Int
        Defines how often the model should be saved (defined relative to the number of refining iters)
    start_episode: Int
        the episode index for tensorflow logging purposes
    exploration_noise: bool
            Defines if exploration noise should be added to the action to be executed.
    log_dir: string
        Defines the log directory to save the normalization statistics in.
    tf_writer: tf.summary
            Tensorflow writer to be used in logging the data.

    Returns
    -------
    system_dynamics_handler: SystemDynamicsHandler
        The system_dynamics_handler holds the trained system dynamics.
    mpc_policy: ModelBasedBasePolicy
        The policy that was refined to be used as a control policy
    """
    if number_of_initial_rollouts > 0:
        system_dynamics_handler = learn_dynamics_from_policy(
            env=env,
            policy=initial_policy,
            number_of_rollouts=number_of_initial_rollouts,
            task_horizon=task_horizon,
            dynamics_function=dynamics_function,
            system_dynamics_handler=system_dynamics_handler,
            epochs=epochs,
            learning_rate=learning_rate,
            validation_split=validation_split,
            batch_size=batch_size,
            is_normalized=is_normalized,
            nn_optimizer=nn_optimizer,
            tf_writer=tf_writer,
            exploration_noise=exploration_noise,
            log_dir=log_dir,
            save_model_frequency=save_model_frequency,
            saved_model_dir=saved_model_dir)
        logging.info("Trained initial system model")
    else:
        if system_dynamics_handler is None:
            system_dynamics_handler = SystemDynamicsHandler(
                env_action_space=env_action_space,
                env_observation_space=env_observation_space,
                true_model=False,
                dynamics_function=dynamics_function,
                tf_writer=tf_writer,
                is_normalized=is_normalized,
                log_dir=log_dir,
                save_model_frequency=save_model_frequency,
                saved_model_dir=saved_model_dir)
    if refinement_policy is None:
        refinement_policy = MPCPolicy(reward_function=reward_function,
                               env_action_space=env_action_space,
                               env_observation_space=env_observation_space,
                               dynamics_handler=system_dynamics_handler,
                               optimizer=optimizer,
                               optimizer_name=optimizer_name,
                               num_agents=num_agents,
                               planning_horizon=planning_horizon,
                               tf_writer=tf_writer,
                               **optimizer_args)
    for i in range(number_of_refinement_steps):
        system_dynamics_handler = learn_dynamics_from_policy(
            env=env,
            policy=refinement_policy,
            number_of_rollouts=number_of_rollouts_for_refinement,
            task_horizon=task_horizon,
            system_dynamics_handler=system_dynamics_handler,
            epochs=epochs,
            learning_rate=learning_rate,
            validation_split=validation_split,
            batch_size=batch_size,
            is_normalized=is_normalized,
            nn_optimizer=nn_optimizer,
            tf_writer=tf_writer,
            exploration_noise=exploration_noise,
            start_episode=start_episode + (number_of_rollouts_for_refinement*i))
    return system_dynamics_handler, refinement_policy



import tensorflow as tf
import numpy as np


@tf.function
def _pendulum_angle_normalize(x):
    return (((x + np.pi) % (2 * np.pi)) - np.pi)


@tf.function
def pendulum_reward_function(current_state, next_state, actions):
    """
        The pendulum state reward function

       Parameters
       ---------
       current_state: tf.float32
            represents the current state of the system (Bxdim_S)
       next_state: tf.float32
            represents the next state of the system (Bxdim_S)

       Returns
        -------
        rewards: tf.float32
            The reward corresponding to each of the pairs current_state, next_state
       """
    return -((_pendulum_angle_normalize(
        tf.math.atan2(current_state[:, 1], current_state[:, 0]))
              ** tf.constant(2, dtype=tf.float32)) + tf.constant(0.1,
                                                                 dtype=tf.float32) * current_state[
                                                                                     :,
                                                                                     2]
             ** tf.constant(2, dtype=tf.float32)) - \
           (tf.constant(0.001, dtype=tf.float32) *
            tf.reduce_sum(tf.square(actions), axis=1))


class PendulumTrueModel(tf.Module):
    def __init__(self, name=None):
        """
        This is the pendulum true model for the gym environment


        Parameters
        ---------
        name: String
            Defines the name of the block of the pendulum true model.
        """
        super(PendulumTrueModel, self).__init__(name=name)
        self.g = tf.constant(10, dtype=tf.float32)
        self.max_torque = tf.constant(2.0, dtype=tf.float32)
        self.max_speed = tf.constant(8.0, dtype=tf.float32)
        self.m = tf.constant(1., dtype=tf.float32)
        self.l = tf.constant(1., dtype=tf.float32)
        self.dt = tf.constant(.05, dtype=tf.float32)
        self.pi = tf.constant(float(np.pi), dtype=tf.float32)

    @tf.function
    def __call__(self, x, train): #cos(theta), sin(theta), dtheta, u
        """
        This is the call function for the pendulum true model.


        Parameters
        ---------
        x: tf.float32
            Defines the (s_t, a_t) which is the state and action stacked on top of each other,
            (dims = Batch X (dim_S + dim_U)) [cos(theta), sin(theta), dtheta, u]
        train: tf.bool
            Placeholder to confirm with the base class.


        Returns
        -------
        output: tf.float32
            Defines the next state (s_t+1) with (dims = Batch X dim_S), [cos(theta), sin(theta), dtheta]
        """
        u = x[:, 3]
        thdot = x[:, 2]
        theta_cos = x[:, 0]
        theta_sin = x[:, 1]
        theta = tf.math.atan2(theta_sin, theta_cos)
        newthdot = thdot + (-tf.constant(3, dtype=tf.float32)*self.g/ (tf.constant(2, dtype=tf.float32)*self.l)
                            * tf.math.sin(theta + self.pi) + tf.constant(3, dtype=tf.float32) /
                            (self.m * self.l ** tf.constant(2, dtype=tf.float32)) * u) * self.dt
        newth = theta + newthdot * self.dt
        newthdot = tf.clip_by_value(newthdot, -self.max_speed, self.max_speed)
        new_state = tf.concat([tf.expand_dims(tf.math.cos(newth), -1),
                          tf.expand_dims(tf.math.sin(newth), -1),
                          tf.expand_dims(newthdot, -1)], axis=1)
        deviation = new_state - x[:, :3]
        return deviation



from gym.wrappers.monitoring.video_recorder import VideoRecorder
from blackbox_mpc.policies.model_free_base_policy import ModelFreeBasePolicy


def record_rollout(env, horizon, policy, record_file_path):
    """
    This is the recording function for the runner class which samples one episode with a specified length
    using the provided policy and records it in a video.


    Parameters
    ---------
    horizon: Int
        The task horizon/ episode length.
    policy: ModelBasedBasePolicy or ModelFreeBasePolicy
        The policy to be used in collecting the episodes from the different agents.
    record_file_path: String
        specified the file path to save the video that will be recorded in.
    """
    recorder = VideoRecorder(env,
                             record_file_path + '.mp4')
    observations = env.reset()
    for t in range(horizon):
        recorder.capture_frame()
        if not isinstance(policy, ModelFreeBasePolicy):
            action_to_execute, expected_obs, expected_reward = policy.act(
                observations, t)
        else:
            action_to_execute = policy.act(observations, t)
        observations, reward, done, info = env.step(action_to_execute)
    recorder.capture_frame()
    recorder.close()
    return



import gym
from blackbox_mpc.environment_utils.subprocess_env import SubprocVecEnv
import multiprocessing as mp


class EnvironmentWrapper:
    @staticmethod
    def make_standard_gym_env(env_name, random_seed=0, num_of_agents=1):
        """
        This is the make env function for standard gym envs which is responsible
        for creating the parallel environment. This takes care of traditional gym
        and mujoco envs.

        Parameters
        ---------
        env_name: String
            This specifies the standard env gym name.
        random_seed: Int
            This specifies the seed to use.
        num_of_agents: Int
            This specifies the number of agents to use.

        Returns
        -------
        env: SubprocVecEnv
            A parellel environment for n agents running in parellel.
        """
        def make_envs(env_name_sub, rank):
            def _make_envs():
                env = gym.make(env_name_sub)
                # env = gym.wrappers.TimeLimit(env, max_time_steps)
                env.seed(random_seed + rank)
                return env

            return _make_envs

        queue = mp.Queue()
        env = SubprocVecEnv([make_envs(env_name, i) for i in range(num_of_agents)],
                             queue=queue)
        return env

    @staticmethod
    def make_custom_gym_env(env_class, random_seed=0, num_of_agents=1):
        """
           This is the make env function for custom gym envs which is responsible
           for creating the parallel environment.
           This takes care of custom gym and mujoco envs.

           Parameters
           ---------
           env_class: gymEnv
               This specifies the class to be used in instantiating the envs.
           random_seed: Int
               This specifies the seed to use.
           num_of_agents: Int
               This specifies the number of agents to use.

           Returns
           -------
           env: SubprocVecEnv
                A parellel environment for n agents running in parellel.
           """

        def make_envs(rank):
            def _make_envs():
                env = env_class()
                # env = gym.wrappers.TimeLimit(env, max_time_steps)
                env.seed(random_seed + rank)
                return env

            return _make_envs

        queue = mp.Queue()
        env = SubprocVecEnv([make_envs(i) for i in range(num_of_agents)],
                             queue=queue)
        return env






import tensorflow as tf


class EvaluatorBase(tf.Module):
    """This is the base class of the trajectory evaluators"""
    def __init__(self, reward_function,
                 system_dynamics_handler,
                 name=None):
        """
        This is the initializer function for the Evaluator Base Class.


        Parameters
        ---------
        name: String
            Defines the name of the block of the evaluator.
        """
        super(EvaluatorBase, self).__init__(name=name)
        self._reward_function = reward_function
        self._system_dynamics_handler = system_dynamics_handler

    @tf.function
    def __call__(self, current_states, action_sequences, time_step):
        """
          This is the call function for the Evaluator Base Class.
          It is used to calculate the rewards corresponding to each of the action sequences starting
          from the current state.

          Parameters
          ---------
          current_states: tf.float32
              Defines the current state of the system, (dims=num_of_agents X dim_S)
          action_sequences: tf.float32
             Defines the action sequences to be evaluated, (dims = population X num_of_agents X planning_horizon X dim_U)
          time_step: tf.float32
              Defines the current timestep of the episode.


          Returns
          -------
          rewards: tf.float32
              The rewards corresponding to each action sequence (dims = 1 X population)
          """
        raise Exception("__call__ function is not implemented yet")

    def predict_next_state(self, current_state, current_action):
        """
          This is the function used to predict the next state using the internal dynamics handler.

          Parameters
          ---------
          current_state: tf.float32
              Defines the current state of the system, (dims=num_of_agents X dim_S)
          current_action: tf.float32
             Defines the current action to be applied, (dims = num_of_agents X dim_U)


          Returns
          -------
          next_state: tf.float32
              Defines the next state of the system, (dims=num_of_agents X dim_S)
          """
        raise Exception("predict_next_state function is not implemented yet")

    def evaluate_next_reward(self, current_state, next_state, current_action):
        """
          This is the function used to predict the next reward using the internal dynamics handler.

          Parameters
          ---------
          current_state: tf.float32
              Defines the current state of the system, (dims=num_of_agents X dim_S)
          next_state: tf.float32
              Defines the next state of the system, (dims=num_of_agents X dim_S)
          current_action: tf.float32
             Defines the current action to be applied, (dims = num_of_agents X dim_U)


          Returns
          -------
          reward: tf.float32
              returns the predicted reward using the action, current state and the next one,
              (dims=num_of_agents X 1)
          """
        raise Exception("evaluate_next_reward function is not implemented yet")



import tensorflow as tf


class DeterministicMLP(tf.Module):
    def __init__(self, layers, activation_functions,
                 loss_fn=tf.keras.losses.MeanSquaredError(),
                 name=None):
        """
        A deterministic fully connected MLP dynamics function class for
        (s_t, a_t) - > (s_t+1)

        Parameters
        ---------
        name: str
            Defines the name of the block of the deterministic MLP dynamics
            function.
        """
        super(DeterministicMLP, self).__init__(name=name)
        self.layers = []
        for i in range(1, len(layers)):
            self.layers.append(
                tf.keras.layers.Dense(layers[i], input_shape=(layers[i-1],),
                                      activation=activation_functions[i-1],
                                      dtype='float32'))
        self.loss_fn = loss_fn

    @tf.function
    def __call__(self, x, train):
        """
        This is the call function for the deterministic fully connected MLP
        dynamics function class.


        Parameters
        ---------
        x: tf.float32
            Defines the (s_t, a_t) which is the state and action stacked on top
            of each other, (dims = Batch X (dim_S + dim_U))

        train: tf.bool
            Defines whether the network should run in train mode or not.


        Returns
        -------
        output: tf.float32
            Defines the next state (s_t+1) with (dims = Batch X dim_S)
        """
        for layer in self.layers.layers:
            x = layer(x)
        return x

    @tf.function
    def get_loss(self, expected_output, predictions):
        """
        This is the training loss function for the deterministic fully
        connected MLP dynamics function class.


        Parameters
        ---------
        expected_output: tf.float32
            Defines the next state (s_t+1) ground truth with (dims = Batch X dim_S)
        predictions: tf.float32
            Defines the next state (s_t+1) predicted values with (dims = Batch X dim_S)


        Returns
        -------
        train_loss: tf.float32
            Defines the training loss as a scalar for the whole batch
        """
        return self.loss_fn(expected_output, predictions)

    @tf.function
    def get_validation_loss(self, expected_output, predictions):
        """
        This is the validation loss function for the deterministic fully
        connected MLP dynamics function class.


        Parameters
        ---------
        expected_output: tf.float32
            Defines the next state (s_t+1) ground truth with (dims = Batch X dim_S)
        predictions: tf.float32
            Defines the next state (s_t+1) predicted values with (dims = Batch X dim_S)


        Returns
        -------
        validation_loss: tf.float32
            Defines the validation loss as a scalar for the whole batch
        """
        return self.loss_fn(expected_output, predictions)



import tensorflow as tf
import numpy as np
from blackbox_mpc.optimizers.optimizer_base import OptimizerBase


class CMAESOptimizer(OptimizerBase):
    def __init__(self, env_action_space, env_observation_space,
                 planning_horizon=50, max_iterations=5, population_size=500,
                 num_elite=50, num_agents=5, alpha_cov=tf.constant(2.0, dtype=tf.float32),
                 h_sigma=tf.constant(1.0, dtype=tf.float32)):
        """
        This class defines a Covariance Matrix Adaptation Evolutionary-Strategy.
        (https://arxiv.org/pdf/1604.00772.pdf) Note: this optimzer is not optimized for more than one agent

        Parameters
        ---------
        env_action_space: gym.ActionSpace
            Defines the action space of the gym environment.
        env_observation_space: gym.ObservationSpace
            Defines the observation space of the gym environment.
        planning_horizon: Int
            Defines the planning horizon for the optimizer (how many steps to lookahead and optimize for).
        max_iterations: tf.int32
            Defines the maximimum iterations for the CMAES optimizer to refine its guess for the optimal solution.
        population_size: tf.int32
            Defines the population size of the particles evaluated at each iteration.
        num_elite: tf.int32
            Defines the number of elites kept for the next iteration from the population.
        num_agents: tf.int32
            Defines the number of runner running in parallel
        alpha_cov: tf.float32
            Defines the alpha covariance to be used.
        h_sigma: tf.float32
            Defines the h sigma to be used.
        """
        super(CMAESOptimizer, self).__init__(name=None,
                                             planning_horizon=planning_horizon,
                                             max_iterations=max_iterations,
                                             num_agents=num_agents,
                                             env_action_space=env_action_space,
                                             env_observation_space=
                                             env_observation_space)
        self._solution_dim = [self._num_agents,
                              self._planning_horizon,
                              self._dim_U]
        self._population_size = population_size
        self._num_elite = num_elite
        previous_solution_values = tf.constant(np.tile((self._action_lower_bound +
                                                        self._action_upper_bound) / 2,
                                                       [self._planning_horizon *
                                                        self._num_agents, 1]),
                                               dtype=tf.float32)
        previous_solution_values = tf.reshape(previous_solution_values, [-1])
        solution_variance_values = tf.constant(np.tile(np.square(self._action_lower_bound
                                                                 - self._action_upper_bound) / 16,
                                                       [self._planning_horizon *
                                                        self._num_agents, 1]),
                                               dtype=tf.float32)
        solution_variance_values = tf.reshape(solution_variance_values, [-1])

        # Recombination weights
        self._weights = tf.concat([
            tf.math.log(tf.cast(self._num_elite, dtype=tf.float32) + 0.5) -
            tf.math.log(tf.range(1, tf.cast(self._num_elite, dtype=tf.float32) + 1)),
            tf.zeros(shape=(self._population_size - self._num_elite,), dtype=tf.float32),
        ], axis=0)
        # Normalize weights such as they sum to one and reshape into a column matrix
        self._weights = (self._weights / tf.reduce_sum(self._weights))[:, tf.newaxis]
        self._mu_eff = tf.reduce_sum(self._weights) ** 2 / \
                       tf.reduce_sum(self._weights ** 2)
        self._solution_size = tf.reduce_prod(self._solution_dim)
        #step_size_control
        self._c_sigma = (self._mu_eff + 2) / (tf.cast(self._solution_size,
                                                      dtype=tf.float32) +
                                              self._mu_eff + 5)
        self._d_sigma = 1 + 2 * tf.maximum(0, tf.sqrt((self._mu_eff - 1) /
                                                      (tf.cast(self._solution_size,
                                                               dtype=tf.float32) + 1)) - 1) \
                        + self._c_sigma
        #Covariance Matrix Adaptation
        self._cc = (4 + self._mu_eff / tf.cast(self._solution_size, dtype=tf.float32)) / \
                    (tf.cast(self._solution_size, dtype=tf.float32) + 4 + 2 * self._mu_eff /
                    tf.cast(self._solution_size, dtype=tf.float32))
        self._alpha_cov = alpha_cov
        self._h_sigma = h_sigma
        self._c1 = self._alpha_cov / ((tf.cast(self._solution_size,
                                               dtype=tf.float32) + 1.3) ** 2 +
                                      self._mu_eff)
        c_mu_option_two = self._alpha_cov * (self._mu_eff - 2 + 1 / self._mu_eff) / \
                          ((tf.cast(self._solution_size, dtype=tf.float32) + 2)
                           ** 2 + self._alpha_cov * self._mu_eff / 2)
        self._c_mu = tf.minimum(1 - self._c1, c_mu_option_two)
        #define trainable parameters
        # Mean
        self._m = tf.Variable(previous_solution_values)
        # Step-size
        self._sigma = tf.Variable(tf.math.sqrt(solution_variance_values))
        # Covariance matrix
        self._C = tf.Variable(tf.eye(num_rows=tf.cast(self._solution_size,
                                                      dtype=tf.float32),
                                     dtype=tf.float32))
        # Evolution path for σ
        self._p_sigma = tf.Variable(tf.zeros((tf.cast(self._solution_size,
                                                      dtype=tf.float32),),
                                             dtype=tf.float32))
        # Evolution path for C
        self._p_C = tf.Variable(tf.zeros((tf.cast(self._solution_size,
                                                  dtype=tf.float32),),
                                         dtype=tf.float32))
        # Coordinate system (normalized eigenvectors)
        self._B = tf.Variable(tf.eye(num_rows=tf.cast(self._solution_size,
                                                      dtype=tf.float32),
                                     dtype=tf.float32))
        # Scaling (square root of eigenvalues)
        self._D = tf.Variable(tf.eye(num_rows=tf.cast(self._solution_size,
                                                      dtype=tf.float32),
                                     dtype=tf.float32))
        self._expectation_of_normal = tf.sqrt(tf.cast(self._solution_size,
                                                      dtype=tf.float32) *
                                              (1 - 1 / (4 * tf.cast(
                                                  self._solution_size,
                                                  dtype=tf.float32)) +
                                              1 / (21 * tf.cast(
                                                          self._solution_size,
                                                          dtype=tf.float32)
                                                   ** 2)))
        return

    @tf.function
    def _optimize(self, current_state, time_step):
        def continue_condition(t, mean):
            result = tf.less(t, self._max_iterations)
            return result

        def iterate(t, mean):
            # -----------------------------------------------------
            # (1) Sample a new population of solutions ∼ N(m, σ²C)
            # -----------------------------------------------------
            z = tf.random.normal([self._population_size, self._solution_size], dtype=tf.float32)
            y = tf.matmul(z, tf.matmul(self._B, self._D))
            samples = self._m + self._sigma * y
            samples = tf.reshape(samples, [self._population_size, *self._solution_dim])
            # -------------------------------------------------
            # (2) Selection and Recombination: Moving the Mean
            # -------------------------------------------------
            # Evaluate and sort solutions
            samples_feasible = tf.clip_by_value(samples, self._action_lower_bound_horizon,
                                                self._action_upper_bound_horizon)
            penalty = tf.norm(tf.reshape(samples - samples_feasible,
                                         [self._population_size, self._num_agents, -1]),
                              axis=2) ** 2
            samples = samples_feasible
            # -------------------------------------------------
            # (2) Selection and Recombination: Moving the Mean
            # -------------------------------------------------
            # Evaluate and sort solutions
            rewards = self._trajectory_evaluator(current_state, samples, time_step) - penalty
            rewards = tf.reduce_sum(rewards, axis=1) #TODO: double check this, very flaky
            self._x_sorted = tf.gather(samples, tf.argsort(rewards, direction='DESCENDING'))
            # The new mean is a weighted average of the top-μ solutions
            x_diff = (tf.reshape(self._x_sorted, [self._population_size, self._solution_size]) - self._m)
            x_mean = tf.reduce_sum(tf.multiply(x_diff, self._weights), axis=0)
            m = self._m + x_mean
            # ----------------------
            # (3) Step-size control
            # ----------------------
            y_mean = x_mean / self._sigma
            D_inv = tf.linalg.tensor_diag(tf.math.reciprocal(tf.linalg.diag_part(self._D)))
            C_inv_half = tf.matmul(tf.matmul(self._B, D_inv), tf.transpose(self._B))
            p_sigma = ((1 - self._c_sigma) * self._p_sigma) + (tf.math.sqrt(self._c_sigma * (2 - self._c_sigma) * self._mu_eff) *
                                                               tf.squeeze(tf.matmul(C_inv_half, y_mean[:, tf.newaxis])))
            sigma = self._sigma * tf.exp((self._c_sigma / self._d_sigma) * ((tf.norm(p_sigma) /
                                                                             self._expectation_of_normal) - 1))
            # -----------------------------------
            # (4) Adapting the Covariance Matrix
            # -----------------------------------
            p_C = ((1 - self._cc) * self._p_C + (self._h_sigma * tf.sqrt(self._cc * (2 - self._cc) * self._mu_eff) * y_mean))

            p_C_matrix = p_C[:, tf.newaxis]
            y_mean_unweighted = x_diff / self._sigma
            y_mean_unweighted_squared = tf.map_fn(fn=lambda e: e * tf.transpose(e), elems=y_mean_unweighted[:, tf.newaxis])
            y_s = tf.reduce_sum(tf.multiply(y_mean_unweighted_squared, self._weights[:, tf.newaxis]), axis=0)
            C = ((1 - self._c1 - self._c_mu) * self._C + self._c1 * p_C_matrix * tf.transpose(p_C_matrix) +
                 self._c_mu * y_s)
            # -----------------------------------
            # (5) Ensure the symmetry of the covariance matrix here
            # -----------------------------------
            C_upper = tf.linalg.band_part(C, 0, -1)
            C_upper_no_diag = C_upper - tf.linalg.tensor_diag(tf.linalg.diag_part(C_upper))
            C = C_upper + tf.transpose(C_upper_no_diag)

            # -----------------------------------
            # (6)Update the values
            # -----------------------------------
            u, B, _ = tf.linalg.svd(C)

            diag_D = tf.sqrt(u)
            D = tf.linalg.tensor_diag(diag_D)
            # Assign values
            self._p_C.assign(p_C)
            self._p_sigma.assign(p_sigma)
            self._C.assign(C)
            self._sigma.assign(sigma)
            self._B.assign(B)
            self._D.assign(D)
            self._m.assign(m)
            return t + tf.constant(1, dtype=tf.int32), m

        _ = tf.while_loop(cond=continue_condition, body=iterate, loop_vars=[tf.constant(0, dtype=tf.int32),
                                                                            self._m])
        solution = tf.reshape(self._m, self._solution_dim)
        resulting_action = solution[:, 0]
        return resulting_action

    def reset(self):
        """
          This method resets the optimizer to its default state at the beginning of the trajectory/episode.
          """
        previous_solution_values = tf.constant(np.tile((self._action_lower_bound + self._action_upper_bound) / 2,
                                                       [self._planning_horizon * self._num_agents, 1]), dtype=tf.float32)
        previous_solution_values = tf.reshape(previous_solution_values, [-1])
        solution_variance_values = tf.constant(
            np.tile(np.square(self._action_lower_bound - self._action_upper_bound) / 16,
                    [self._planning_horizon * self._num_agents, 1]), dtype=tf.float32)
        solution_variance_values = tf.reshape(solution_variance_values, [-1])
        self._m.assign(previous_solution_values)
        self._sigma.assign(tf.math.sqrt(solution_variance_values))



import tensorflow as tf
import numpy as np
from blackbox_mpc.optimizers.optimizer_base import OptimizerBase


class SPSAOptimizer(OptimizerBase):
    def __init__(self, env_action_space, env_observation_space,
                 planning_horizon=50, max_iterations=5, population_size=500,
                 num_agents=5, alpha=tf.constant(0.602, dtype=tf.float32),
                 gamma=tf.constant(0.101, dtype=tf.float32),
                 a_par=tf.constant(0.01, dtype=tf.float32),
                 noise_parameter=tf.constant(0.3, dtype=tf.float32)):
        """
           This class defines the simultaneous perturbation stochastic approximation optimizer.
           (https://www.jhuapl.edu/SPSA/PDF-SPSA/Spall_Stochastic_Optimization.PDF)


           Parameters
           ---------
           env_action_space: gym.ActionSpace
                Defines the action space of the gym environment.
           env_observation_space: gym.ObservationSpace
                Defines the observation space of the gym environment.
           planning_horizon: Int
               Defines the planning horizon for the optimizer (how many steps to lookahead and optimize for).
           max_iterations: tf.int32
               Defines the maximimum iterations for the CMAES optimizer to refine its guess for the optimal solution.
           population_size: tf.int32
               Defines the population size of the particles evaluated at each iteration.
           num_agents: tf.int32
               Defines the number of runner running in parallel
           alpha: tf.float32
               Defines the alpha used.
           gamma: tf.float32
               Defines the gamma used.
           a_par: tf.float32
               Defines the a_par used.
           noise_parameter: tf.float32
               Defines the noise_parameter used.
        """
        super(SPSAOptimizer, self).__init__(name=None,
                                           planning_horizon=planning_horizon,
                                           max_iterations=max_iterations,
                                           num_agents=num_agents,
                                           env_action_space=env_action_space,
                                           env_observation_space=
                                           env_observation_space)
        self._solution_dim = [self._num_agents, self._planning_horizon, self._dim_U]
        self._population_size = population_size
        current_params = np.tile((self._action_lower_bound + self._action_upper_bound) / 2,
                                 [self._planning_horizon * self._num_agents, 1])
        current_params = current_params.reshape([self._num_agents, self._planning_horizon, -1])
        self._alpha = alpha
        self._gamma = gamma
        self._a_par = a_par
        self._big_a_par = tf.cast(self._max_iterations, dtype=tf.float32) / tf.constant(10., dtype=tf.float32)
        self._noise_parameter = noise_parameter
        self._current_parameters = tf.Variable(tf.constant(current_params,
                                                           dtype=tf.float32))

    @tf.function
    def _optimize(self, current_state, time_step):
        def continue_condition(t, solution):
            result = tf.less(t, self._max_iterations)
            return result

        def iterate(t, solution):
            #TODO: early termination and checking if we are always doing better or we diverged
            ak = self._a_par / (tf.cast(t, tf.float32) + 1 + self._big_a_par) ** self._alpha
            ck = self._noise_parameter / (tf.cast(t, tf.float32) + 1) ** self._gamma

            #_sample delta for half of the population
            delta = (tf.random.uniform(shape=[self._population_size, *self._solution_dim],
                                       minval=0, maxval=2, dtype=tf.int32)*2) - 1
            delta = tf.cast(delta, dtype=tf.float32)

            parameters_plus = solution + ck * delta
            parameters_minus = solution - ck * delta

            parameters_plus_feasible = tf.clip_by_value(parameters_plus, self._action_lower_bound_horizon,
                                                        self._action_upper_bound_horizon)
            parameters_minus_feasible = tf.clip_by_value(parameters_minus, self._action_lower_bound_horizon,
                                                         self._action_upper_bound_horizon)
            parameters_plus_penalty = tf.norm(tf.reshape(parameters_plus - parameters_plus_feasible,
                                                         [self._population_size, self._num_agents, -1]),
                                              axis=2) ** 2
            parameters_minus_penalty = tf.norm(tf.reshape(parameters_minus - parameters_minus_feasible,
                                                          [self._population_size, self._num_agents, -1]),
                                               axis=2) ** 2
            parameters_plus = parameters_plus_feasible
            parameters_minus = parameters_minus_feasible

            #concat both for faster implementation
            actions_inputs = tf.concat([parameters_plus, parameters_minus], axis=0)
            full_rewards = self._trajectory_evaluator(current_state, actions_inputs,
                                                      time_step)
            #evaluate the costs
            rewards_parameters_plus = full_rewards[0:self._population_size] - parameters_plus_penalty
            rewards_parameters_minus = full_rewards[self._population_size:] - parameters_minus_penalty
            # Estimate the gradient
            ghat = tf.reduce_mean(tf.expand_dims(tf.expand_dims(rewards_parameters_plus -
                                                               rewards_parameters_minus, -1), -1) / (2. * ck * delta),
                                  axis=0)
            #update now the parameters
            new_solution = solution + ak * ghat
            new_solution = tf.clip_by_value(new_solution, self._action_lower_bound_horizon,
                                            self._action_upper_bound_horizon)

            return t + tf.constant(1, dtype=tf.int32), new_solution

        _, solution = tf.while_loop(cond=continue_condition, body=iterate,
                                    loop_vars=[tf.constant(0, dtype=tf.int32), self._current_parameters])
        #shift the solution for next iteration start
        self._current_parameters.assign(tf.concat([solution[:, 1:],
                                                   tf.expand_dims(solution[:, -1], 1)], 1))
        resulting_action = solution[:, 0]
        return resulting_action

    def reset(self):
        """
         This method resets the optimizer to its default state at the beginning of the trajectory/episode.
         """
        current_params = np.tile((self._action_lower_bound + self._action_upper_bound) / 2,
                                 [self._planning_horizon * self._num_agents, 1])
        current_params = current_params.reshape([self._num_agents, self._planning_horizon, -1])
        self._current_parameters.assign(current_params)
        return



import tensorflow as tf
import numpy as np
import tensorflow_probability as tfp
from blackbox_mpc.optimizers.optimizer_base import OptimizerBase
tfd = tfp.distributions


class PI2Optimizer(OptimizerBase):
    def __init__(self, env_action_space, env_observation_space,
                 planning_horizon=50, max_iterations=5, population_size=500,
                 num_agents=5, lamda=tf.constant(1.0, dtype=tf.float32)):
        """
        This class defines the information theortic MPC based on path intergral methods.
        (https://ieeexplore.ieee.org/stamp/stamp.jsp?tp=&arnumber=7989202)


        Parameters
        ---------
        env_action_space: gym.ActionSpace
            Defines the action space of the gym environment.
        env_observation_space: gym.ObservationSpace
            Defines the observation space of the gym environment.
        planning_horizon: Int
            Defines the planning horizon for the optimizer (how many steps to lookahead and optimize for).
        max_iterations: tf.int32
            Defines the maximimum iterations for the CMAES optimizer to refine its guess for the optimal solution.
        population_size: tf.int32
            Defines the population size of the particles evaluated at each iteration.
        num_agents: tf.int32
            Defines the number of runner running in parallel
        lamda: tf.float32
            Defines the lamda used the energy function.
        """
        super(PI2Optimizer, self).__init__(name=None,
                                           planning_horizon=planning_horizon,
                                           max_iterations=max_iterations,
                                           num_agents=num_agents,
                                           env_action_space=env_action_space,
                                           env_observation_space=
                                           env_observation_space)
        self._solution_dim = [self._num_agents, self._planning_horizon, self._dim_U]
        self._solution_size = tf.reduce_prod(self._solution_dim)
        self._population_size = population_size
        previous_solution_values = np.tile((self._action_lower_bound + self._action_upper_bound) / 2,
                                           [self._planning_horizon * self._num_agents, 1])
        previous_solution_values = previous_solution_values.reshape([self._num_agents, self._planning_horizon, -1])
        self._previous_solution = tf.Variable(tf.zeros(shape=previous_solution_values.shape,
                                                       dtype=tf.float32))
        self._previous_solution.assign(previous_solution_values)
        solution_variance_values = np.tile(np.square(self._action_lower_bound - self._action_upper_bound) / 16,
                                           [self._planning_horizon * self._num_agents, 1])
        solution_variance_values = solution_variance_values.reshape([self._num_agents, self._planning_horizon, -1])
        self._solution_variance = tf.Variable(tf.zeros(shape=solution_variance_values.shape,
                                                       dtype=tf.float32))
        self._solution_variance.assign(solution_variance_values)
        self._lamda = lamda

    @tf.function
    def _optimize(self, current_state, time_step):
        def continue_condition(t, mean):
            result = tf.less(t, self._max_iterations)
            return result

        def iterate(t, mean):
            samples = tf.random.truncated_normal([self._population_size,
                                                  *self._solution_dim],
                                                 mean,
                                                 tf.sqrt(self._solution_variance),
                                                 dtype=tf.float32)
            samples_feasible = tf.clip_by_value(samples, self._action_lower_bound_horizon,
                                                self._action_upper_bound_horizon)
            penalty = tf.norm(tf.reshape(samples - samples_feasible, [self._population_size,
                                                                      self._num_agents,
                                                                      -1]),
                              axis=2) ** 2
            samples = samples_feasible

            rewards = self._trajectory_evaluator(current_state, samples, time_step) - penalty
            costs = -rewards
            costs = tf.transpose(costs, [1, 0])
            beta = tf.reduce_min(costs, axis=1)
            prob = tf.math.exp(-(1 / self._lamda) * (costs - tf.expand_dims(beta, -1)))
            eta = tf.reduce_sum(prob, axis=1)
            #compute weights now
            omega = tf.expand_dims(1 / eta, -1) * prob
            samples = tf.transpose(samples, [1, 0, 2, 3])
            new_mean = tf.reduce_sum(tf.multiply(samples, tf.expand_dims(tf.expand_dims(omega, -1), -1)), axis=1)
            return t + tf.constant(1, dtype=tf.int32), new_mean

        _, new_mean = tf.while_loop(cond=continue_condition, body=iterate, loop_vars=[tf.constant(0, dtype=tf.int32),
                                                                                      self._previous_solution])
        self._previous_solution.assign(tf.concat([new_mean[:, 1:],
                                                  tf.expand_dims(new_mean[:, -1], 1)], 1))

        resulting_action = new_mean[:, 0]
        return resulting_action

    def reset(self):
        """
          This method resets the optimizer to its default state at the beginning of the trajectory/episode.
          """
        previous_solution_values = np.tile((self._action_lower_bound + self._action_upper_bound) / 2,
                                           [self._planning_horizon * self._num_agents, 1])
        previous_solution_values = previous_solution_values.reshape([self._num_agents, self._planning_horizon, -1])
        self._previous_solution.assign(previous_solution_values)



import logging
logging.getLogger().setLevel(logging.INFO)
from blackbox_mpc.policies.model_free_base_policy import ModelFreeBasePolicy
import time
import tensorflow as tf
import numpy as np
from blackbox_mpc.policies.random_policy import RandomPolicy


def perform_rollouts(env, number_of_rollouts, task_horizon, policy,
                     exploration_noise=False, tf_writer=None,
                     start_episode=0):
    """
    This is the perform_rollouts function for the runner class which samples n episodes with a specified length
    using the provided policy.


    Parameters
    ---------
    env: parallelgymEnv
        a wrapped gym environment using blackbox.environment_utils.EnvironmentWrapper funcs
    number_of_rollouts: Int
        Number of rollouts/ episodes to perform for each of the agents in the vectorized environment.
    task_horizon: Int
        The task horizon/ episode length.
    policy: ModelBasedBasePolicy or ModelFreeBasePolicy
        The policy to be used in collecting the episodes from the different agents.
    exploration_noise: bool
        If noise should be added to the actions to help in exploration.
    tf_writer: tf.summary
            Tensorflow writer to be used in logging the data.
    start_episode: Int
        the episode index for tensorflow logging purposes

    Returns
    -------
    traj_obs: [np.float32]
        List with length=number_of_rollouts which holds the observations starting from the reset observations.
    traj_acs: [np.float32]
        List with length=number_of_rollouts which holds the actions taken by the policy.
    traj_rews: [np.float32]
        List with length=number_of_rollouts which holds the rewards taken by the policy.
    """
    traj_obs, traj_acs, traj_rews = [], [], []
    samples = []
    logging.info("Started collecting samples for rollouts")
    for i in range(number_of_rollouts):
        samples.append(
            _sample(
                env, task_horizon, policy, exploration_noise=exploration_noise,
                tf_writer=tf_writer,
                episode_step=start_episode+i))
        traj_obs.append(samples[-1]["observations"])
        traj_acs.append(samples[-1]["actions"])
        traj_rews.append(samples[-1]["rewards"])
    logging.info("Finished collecting samples for rollout")
    return traj_obs, traj_acs, traj_rews


def _sample(env, horizon, policy, episode_step,
            exploration_noise=False, tf_writer=None):
    """
    This is the sampling function for the runner class which samples one episode with a specified length
    using the provided policy.


    Parameters
    ---------
    env: parallelgymEnv
        a wrapped gym environment using blackbox.environment_utils.EnvironmentWrapper funcs
    horizon: Int
        The task horizon/ episode length.
    policy: ModelBasedBasePolicy or ModelFreeBasePolicy
        The policy to be used in collecting the episodes from the different agents.
    exploration_noise: bool
        If noise should be added to the actions to help in exploration.

    Returns
    -------
    result: dict
        returns the episode rollouts results for all the agents in the parallelized environment,
        it has the form of {observations, actions, rewards, reward_sum}
    """
    policy.reset()
    first_obs = env.reset()
    times, observations, actions, rewards, reward_sum, done = \
        [], [first_obs], [], [], 0, False
    if not isinstance(policy, ModelFreeBasePolicy):
        predicted_reward = 0

    for t in range(horizon):
        start = time.time()
        if not isinstance(policy, ModelFreeBasePolicy):
            action_to_execute, expected_obs, expected_reward = \
                policy.act(observations[t], t, exploration_noise)
            predicted_reward += expected_reward
        else:
            action_to_execute = policy.act(observations[t], t)
            action_to_execute = action_to_execute.numpy()
        actions.append(action_to_execute)
        times.append(time.time() - start)
        obs, reward, done, info = env.step(actions[t])
        if tf_writer is not None:
            if not isinstance(policy, RandomPolicy):
                with tf_writer.as_default():
                    tf.summary.scalar('rewards/actual_reward', np.mean(reward),
                                      step=(episode_step*horizon)+t)
            if not isinstance(policy, ModelFreeBasePolicy):
                with tf_writer.as_default():
                    tf.summary.scalar('states/predicted_observations_abs_error',
                                      np.mean(np.sum(np.abs(expected_obs - obs),
                                                     axis=1)),
                                      step=(episode_step*horizon)+t)
                    tf.summary.scalar('rewards/predicted_reward_abs_error',
                                      np.mean(np.abs(expected_reward - reward)),
                                      step=(episode_step * horizon) + t)
        observations.append(obs)
        rewards.append(reward)
        reward_sum += reward
        if t >= horizon - 1:
            if tf_writer is not None:
                if not isinstance(policy, RandomPolicy):
                    with tf_writer.as_default():
                        tf.summary.scalar('rewards/actual_episode_reward',
                                          np.mean(reward_sum),
                                          step=episode_step)
                if not isinstance(policy, ModelFreeBasePolicy):
                    with tf_writer.as_default():
                        tf.summary.scalar('rewards/predicted_episode_reward',
                                          np.mean(predicted_reward),
                                          step=episode_step)
            break
    logging.info("Average action selection time: " + str(np.mean(times)))
    logging.info("Rollout length: " + str(len(actions)))

    return {"observations": np.array(observations),
            "actions": np.array(actions),
            "rewards": np.array(rewards),
            "reward_sum": reward_sum}



import tensorflow as tf
from blackbox_mpc.utils.rollouts import perform_rollouts
from blackbox_mpc.dynamics_handlers.system_dynamics_handler import \
    SystemDynamicsHandler


def learn_dynamics_from_policy(env, policy,
                               number_of_rollouts, task_horizon,
                               dynamics_function=None,
                               system_dynamics_handler=None,
                               epochs=30, learning_rate=1e-3,
                               validation_split=0.2,
                               batch_size=128, is_normalized=True,
                               nn_optimizer=tf.keras.optimizers.Adam,
                               tf_writer=None,
                               exploration_noise=False,
                               log_dir=None,
                               save_model_frequency=1,
                               saved_model_dir=None,
                               start_episode=0):
    """
    This is the learn dynamics function for the runner class which samples n rollouts using a random policy and then
    uses these rollouts to learn a dynamics function for the system.


    Parameters
    ---------
    env: parallelgymEnv
        a wrapped gym environment using blackbox.environment_utils.EnvironmentWrapper funcs
    policy: ModelFreeBasePolicy or ModelBasedBasePolicy
        the policy used for learning the dynamics.
    number_of_rollouts: Int
        Number of rollouts/ episodes to perform for each of the agents in the vectorized environment.
    task_horizon: Int
        The task horizon/ episode length.
    dynamics_function: DeterministicDynamicsFunctionBaseClass
        Defines the system dynamics function.
    learning_rate: float
        Learning rate to be used in training the dynamics function.
    epochs: Int
        Number of epochs to be used in training the dynamics function everytime train is called.
    validation_split: float32
        Defines the validation split to be used of the rollouts collected.
    batch_size: int
        Defines the batch size to be used for training the model.
    nn_optimizer: tf.keras.optimizers
        Defines the optimizer to use with the neural network.
    is_normalized: bool
        Defines if the dynamics function should be trained with normalization or not.
    log_dir: string
        Defines the log directory to save the normalization statistics in.
    tf_writer: tf.summary
            Tensorflow writer to be used in logging the data.
    system_dynamics_handler: SystemDynamicsHandler
            The system_dynamics_handler is a handler of the state, actions and targets processing funcs as well
    saved_model_dir: string
            Defines the saved model directory where the model is saved in, in case of loading the model.
    save_model_frequency: Int
        Defines how often the model should be saved (defined relative to the number of refining iters)
    start_episode: Int
        the episode index for tensorflow logging purposes
    exploration_noise: bool
            Defines if exploration noise should be added to the action to be executed.

    Returns
    -------
    system_dynamics_handler: SystemDynamicsHandler
        The system_dynamics_handler holds the trained system dynamics.
    """
    if system_dynamics_handler is None:
        system_dynamics_handler = SystemDynamicsHandler(env_action_space=env.action_space,
                                                        env_observation_space=env.observation_space,
                                                        true_model=False,
                                                        dynamics_function=dynamics_function,
                                                        tf_writer=tf_writer,
                                                        is_normalized=is_normalized,
                                                        log_dir=log_dir,
                                                        save_model_frequency=save_model_frequency,
                                                        saved_model_dir=saved_model_dir)
    traj_obs, traj_acs, traj_rews = \
        perform_rollouts(env, number_of_rollouts, task_horizon, policy,
                         exploration_noise=exploration_noise,
                         tf_writer=tf_writer,
                         start_episode=start_episode)
    system_dynamics_handler.train(traj_obs, traj_acs, traj_rews,
                                  validation_split=validation_split,
                                  batch_size=batch_size,
                                  learning_rate=learning_rate,
                                  epochs=epochs, nn_optimizer=nn_optimizer)
    return system_dynamics_handler



from blackbox_mpc.policies.model_free_base_policy import ModelFreeBasePolicy
import tensorflow as tf


class RandomPolicy(ModelFreeBasePolicy):
    def __init__(self, number_of_agents, env_action_space):
        """
        This is the random policy for controlling the agent


        Parameters
        ---------
        env_action_space: gym.ActionSpace
            Defines the action space of the gym environment.
        number_of_agents: tf.int32
            Defines the number of runner running in parallel
        """
        super(RandomPolicy, self).__init__()
        self._num_of_agents = number_of_agents
        self._action_lower_bound = tf.constant(env_action_space.high,
                                               dtype=tf.float32)
        self._action_upper_bound = tf.constant(env_action_space.low,
                                               dtype=tf.float32)
        return

    def act(self, observations, t, exploration_noise=False):
        """
        This is the act function for the random policy, which should be called to provide the action
        to be executed at the current time step.


        Parameters
        ---------
        observations: tf.float32
            Defines the current observations received from the environment.
        t: tf.float32
            Defines the current timestep.
        exploration_noise: bool
            Defines if exploration noise should be added to the action to be executed.


        Returns
        -------
        action: tf.float32
            The action to be executed for each of the runner (dims = runner X dim_U)
        """
        return tf.random.uniform([self._num_of_agents, *self._action_lower_bound.shape],
                                 self._action_lower_bound,
                                 self._action_upper_bound, dtype=tf.float32)

    def reset(self):
        """
        This is the reset function for the random policy, which should be called at the beginning of
        the episode.
        """
        return



import tensorflow as tf
import numpy as np
from blackbox_mpc.optimizers.optimizer_base import OptimizerBase


class CEMOptimizer(OptimizerBase):
    def __init__(self, env_action_space, env_observation_space,
                 planning_horizon=50, max_iterations=5, population_size=500,
                 num_elite=50, num_agents=5,
                 epsilon=0.001, alpha=0.25):
        """
        This Class defines a Cross-Entropy Method optimizer.
        (http://web.mit.edu/6.454/www/www_fall_2003/gew/CEtutorial.pdf)


        Parameters
        ---------
        env_action_space: gym.ActionSpace
            Defines the action space of the gym environment.
        env_observation_space: gym.ObservationSpace
            Defines the observation space of the gym environment.
        planning_horizon: Int
            Defines the planning horizon for the optimizer (how many steps to lookahead and optimize for).
        max_iterations: tf.int32
            Defines the maximimum iterations for the CEM optimizer to refine its guess for the optimal solution.
        population_size: tf.int32
            Defines the population size of the particles evaluated at each iteration.
        num_elite: tf.int32
            Defines the number of elites kept for the next iteration from the population.
        num_agents: tf.int32
            Defines the number of runner running in parallel
        epsilon: tf.float32
            Defines the epsilon threshold for the difference between iterations solutions so that
            the optimizer returns the solution earlier than max iterations.
        alpha: tf.float32
            Defines the weight of the solution at t-1 in determining the solution at t,
            ex: mean = alpha*old_mean + (1-alpha)*new_mean.
        """
        super(CEMOptimizer, self).__init__(name=None,
                                           planning_horizon=planning_horizon,
                                           max_iterations=max_iterations,
                                           num_agents=num_agents,
                                           env_action_space=env_action_space,
                                           env_observation_space=
                                           env_observation_space)
        self._solution_dim = [self._num_agents, self._planning_horizon,
                              self._dim_U]
        self._elites_dim = [int(self._num_agents), int(num_elite),
                            int(self._planning_horizon),
                            int(self._dim_U)]
        self._population_size = population_size
        self._num_elite = num_elite
        self._epsilon = epsilon
        self._alpha = alpha
        previous_solution_values = np.tile((self._action_lower_bound +
                                            self._action_upper_bound) / 2,
                                           [self._planning_horizon *
                                            self._num_agents, 1])
        previous_solution_values = previous_solution_values.reshape(
            [self._num_agents, self._planning_horizon, -1])
        self._previous_solution = tf.Variable(
            tf.zeros(shape=previous_solution_values.shape, dtype=tf.float32))
        self._previous_solution.assign(previous_solution_values)
        solution_variance_values = np.tile(np.square(self._action_lower_bound -
                                                     self._action_upper_bound) / 16,
                                           [self._planning_horizon *
                                            self._num_agents, 1])
        solution_variance_values = solution_variance_values.reshape(
            [self._num_agents, self._planning_horizon, -1])
        self._solution_variance = tf.Variable(tf.zeros(
            shape=solution_variance_values.shape, dtype=tf.float32))
        self._solution_variance.assign(solution_variance_values)

    @tf.function
    def _optimize(self, current_state, time_step):
        def continue_condition(t, mean, variance):
            result = tf.less(t, self._max_iterations)
            return result

        def iterate(t, mean, variance):
            lower_bound_dist = mean - self._action_lower_bound_horizon
            upper_bound_dist = self._action_upper_bound_horizon - mean
            constrained_variance = tf.minimum(tf.minimum(
                tf.square(lower_bound_dist /
                          tf.constant(2, dtype=tf.float32)),
                tf.square(upper_bound_dist /
                          tf.constant(2, dtype=tf.float32))),
                                              variance)
            #_sample population
            samples = tf.random.truncated_normal([self._population_size,
                                                  *self._solution_dim],
                                                  mean,
                                                  tf.sqrt(constrained_variance),
                                                 dtype=tf.float32)
            rewards = self._trajectory_evaluator(current_state, samples,
                                                 time_step)
            rewards = tf.transpose(rewards, [1, 0])
            values, indices = tf.nn.top_k(rewards, k=self._num_elite,
                                          sorted=True)
            samples = tf.transpose(samples, [1, 0, 2, 3])
            elites = tf.zeros(self._elites_dim, dtype=tf.float32)
            for agent in tf.range(self._num_agents):
                elites = tf.concat([elites[:agent],
                                    tf.expand_dims(tf.gather(samples[agent],
                                                             indices[agent]), 0),
                                    tf.zeros([samples.shape[0] - 1 - agent,
                                              self._num_elite,
                                              *samples.shape[2:]],
                                             dtype=tf.float32)],
                                   axis=0)
                elites.set_shape(self._elites_dim)
            new_mean = tf.reduce_mean(elites, axis=1)
            new_variance = tf.reduce_mean(tf.square(elites -
                                                    tf.tile(
                                                        tf.expand_dims(
                                                            new_mean, 1),
                                                        [1, tf.shape(elites)
                                                        [1], 1, 1])),
                                          axis=1)

            mean = self._alpha * mean + (tf.constant(
                1, dtype=tf.float32) - self._alpha) * new_mean
            variance = self._alpha * variance + (
                    tf.constant(1, dtype=tf.float32) - self._alpha) * \
                       new_variance

            return t + tf.constant(1, dtype=tf.int32), mean, variance

        num_optimization_iters, mean, variance = tf.while_loop(
            cond=continue_condition, body=iterate,
            loop_vars=[tf.constant(0, dtype=tf.int32), self._previous_solution,
                       self._solution_variance])
        #TODO: the below line is causing problems with cheetah env
        # self.previous_solution.assign(mean)
        resulting_action = mean[:, 0]
        return resulting_action

    def reset(self):
        """
          This method resets the optimizer to its default state at the beginning of the trajectory/episode.
          """
        previous_solution_values = np.tile((self._action_lower_bound +
                                            self._action_upper_bound) / 2,
                                           [self._planning_horizon *
                                            self._num_agents, 1])
        previous_solution_values = \
            previous_solution_values.reshape([self._num_agents,
                                              self._planning_horizon, -1])
        self._previous_solution.assign(previous_solution_values)



import tensorflow as tf
from blackbox_mpc.trajectory_evaluators.evaluator_base import EvaluatorBase


class DeterministicTrajectoryEvaluator(EvaluatorBase):
    def __init__(self, reward_function, system_dynamics_handler):
        """
        This is the trajectory evaluator class for a deterministic dynamics function


        Parameters
        ---------
        reward_function: tf_function
            Defines the reward function with the prototype: tf_func_name(current_state, current_actions, next_state),
            where current_state is BatchXdim_S, next_state is BatchXdim_S and  current_actions is BatchXdim_U.
        system_dynamics_handler: SystemDynamicsHandler
            Defines the system dynamics handler class with its own trainer and observations and actions
             preprocessing functions.
        """
        super(DeterministicTrajectoryEvaluator, self).__init__(
            reward_function=reward_function,
            system_dynamics_handler=system_dynamics_handler,
            name=None)
        return

    @tf.function
    def __call__(self, current_states, action_sequences, time_step):
        """
          This is the call function for the Deterministic Trajectory Evaluator Class.
          It is used to calculate the rewards corresponding to each of the action sequences starting
          from the current state.

          Parameters
          ---------
          current_states: tf.float32
              Defines the current state of the system, (dims=num_of_agents X dim_S)
          action_sequences: tf.float32
             Defines the action sequences to be evaluated, (dims = population X num_of_agents X planning_horizon X dim_U)
          time_step: tf.float32
              Defines the current timestep of the episode.


          Returns
          -------
          rewards: tf.float32
              The rewards corresponding to each action sequence (dims = 1 X population)
          """
        init_t = tf.constant(0, dtype=tf.int32)
        nopt = tf.shape(action_sequences)[0]
        n_agents = tf.shape(action_sequences)[1]
        planning_horizon = tf.shape(action_sequences)[2]
        init_rewards = tf.zeros([nopt*n_agents], dtype=tf.float32)
        action_sequences = tf.reshape(action_sequences,
                                      [-1, planning_horizon,
                                       tf.shape(action_sequences)[3]])
        action_sequences = tf.transpose(action_sequences, [1, 0, 2])
        init_states = tf.tile(current_states, [nopt, 1])

        def continue_prediction(t, total_reward, current_state):
            return tf.less(t, planning_horizon)

        def iterate(t, total_reward, current_state):
            current_actions = action_sequences[t]
            next_state = self.predict_next_state(current_state, current_actions)
            delta_reward = self._reward_function(current_state,
                                                 current_actions, next_state)
            return t + tf.constant(1, dtype=tf.int32), \
                   total_reward + delta_reward, next_state

        _, rewards, _ = tf.while_loop(
            cond=continue_prediction, body=iterate, loop_vars=[init_t, init_rewards,
                                                               init_states]
        )
        rewards = tf.reshape(rewards, [nopt, n_agents])
        return tf.where(tf.math.is_nan(rewards),
                        tf.constant(-1e6, dtype=tf.float32) *
                        tf.ones_like(rewards), rewards)

    @tf.function
    def predict_next_state(self, current_states, current_actions):
        """
          This is the function used to predict the next state using the internal dynamics handler.

          Parameters
          ---------
          current_states: tf.float32
              Defines the current state of the system, (dims=num_of_agents X dim_S)
          current_actions: tf.float32
             Defines the current action to be applied, (dims = num_of_agents X dim_U)


          Returns
          -------
          next_state: tf.float32
              Defines the next state of the system, (dims=num_of_agents X dim_S)
          """
        sys_model_inputs = self._system_dynamics_handler.process_input(
            current_states, current_actions)
        raw_next_states = self._system_dynamics_handler._dynamics_function(
            sys_model_inputs, train=tf.constant(False, dtype=tf.bool))
        next_states = self._system_dynamics_handler.process_output(
            current_states, raw_next_states)
        return next_states

    def evaluate_next_reward(self, current_states,
                             next_states, current_actions):
        """
          This is the function used to predict the next reward using the internal dynamics handler.

          Parameters
          ---------
          current_states: tf.float32
              Defines the current state of the system, (dims=num_of_agents X dim_S)
          next_states: tf.float32
              Defines the next state of the system, (dims=num_of_agents X dim_S)
          current_actions: tf.float32
             Defines the current action to be applied, (dims = num_of_agents X dim_U)


          Returns
          -------
          reward: tf.float32
              returns the predicted reward using the action, current state and the next one,
              (dims=num_of_agents X 1)
          """
        return self._reward_function(current_states, current_actions,
                                     next_states)



import setuptools
setuptools.setup(
     name='blackbox_mpc2',
     version='0.3.0.1',
     author="Ossama Ahmed, Jonas Rothfuss, Ibrahim Ahmed",
     author_email="ibr_alameen@hotmail.com",
     description="Fork of BlackBox MPC - Model Predictive Control with"
                  "sampling based optimizers",
     url="https://github.com/Malborne/blackbox_mpc2.git",
     packages=setuptools.find_packages(),
     install_requires=[
        'tensorflow',
        'tensorflow-probability',
        'gym',
        'numpy',
        'sphinx',
        'matplotlib',
        'sphinx_rtd_theme',
        'sphinxcontrib-bibtex',
      ],
    zip_safe=False
    )


class ModelFreeBasePolicy(object):
    def __init__(self):
        """
            This is the model free policy base class for controlling the agent
        """
        pass

    def act(self, observations, t, exploration_noise=False):
        """
        This is the act function for the model free policy base class,
        which should be called to provide the action
        to be executed at the current time step.


        Parameters
        ---------
        observations: tf.float32
            Defines the current observations received from the environment.
        t: tf.float32
            Defines the current timestep.
        exploration_noise: bool
            Defines if exploration noise should be added to the action that will be executed.
        Returns
        -------
        action: tf.float32
            The action to be executed for each of the runner (dims = runner X dim_U)
        """
        raise Exception("act function is not implemented yet")

    def reset(self):
        """
        This is the reset function for the model free policy base class,
        which should be called at the beginning of
        the episode.
        """
        raise Exception("reset function is not implemented yet")



from blackbox_mpc.policies.model_based_base_policy import ModelBasedBasePolicy
from blackbox_mpc.trajectory_evaluators.deterministic import \
    DeterministicTrajectoryEvaluator
from blackbox_mpc.dynamics_handlers.system_dynamics_handler import \
    SystemDynamicsHandler
import tensorflow as tf
import numpy as np


class MPCPolicy(ModelBasedBasePolicy):
    def __init__(self, trajectory_evaluator=None,
                 optimizer=None, tf_writer=None,
                 log_dir=None, reward_function=None,
                 env_action_space=None, env_observation_space=None,
                 dynamics_function=None, dynamics_handler=None,
                 true_model=False, optimizer_name=None,
                 num_agents=None,
                 save_model_frequency=1,
                 saved_model_dir=None,
                 **optimizer_args):
        """
            This is the model predictive control policy for controlling the agent

        Parameters
        ---------
        trajectory_evaluator: EvaluatorBase
            Defines the trajectory evaluator to be used in the optimizer to
            evaluate trajectories.
        tf_writer: tf.summary
            Tensorflow writer to be used in logging the data.
        optimizer_name: str
            optimizer name between in ['CEM', 'CMA-ES', 'PI2', 'RandomSearch', 'PSO', 'SPSA'].
        env_action_space: gym.ActionSpace
            Defines the action space of the gym environment.
        env_observation_space: gym.ObservationSpace
            Defines the observation space of the gym environment.
        dynamics_function: DeterministicDynamicsFunctionBaseClass
            Defines the system dynamics function.
        dynamics_handler: SystemDynamicsHandler
            The system_dynamics_handler is a handler of the state, actions and targets processing funcs as well
            as the dynamics function.
        reward_function: tf_function
            Defines the reward function with the prototype: tf_func_name(current_state, current_actions, next_state),
            where current_state is BatchXdim_S, next_state is BatchXdim_S and  current_actions is BatchXdim_U.
        true_model: bool
            boolean defining if its a true model dynamics or not.
        log_dir: string
            Defines the log directory to save the normalization statistics in.
        num_agents: tf.int32
            Defines the number of runner running in parallel
        saved_model_dir: string
            Defines the saved model directory where the model is saved in, in case of loading the model.
        save_model_frequency: Int
            Defines how often the model should be saved (defined relative to the number of refining iters)
        optimizer_args: args
            other arguments specific to the optimizer.
        """
        if trajectory_evaluator is None:
            if dynamics_handler is None:
                trajectory_evaluator = \
                    DeterministicTrajectoryEvaluator(reward_function=reward_function,
                                                     system_dynamics_handler=
                        SystemDynamicsHandler(env_action_space=env_action_space,
                                              env_observation_space=env_observation_space,
                                              true_model=true_model,
                                              dynamics_function=dynamics_function,
                                              log_dir=log_dir,
                                              tf_writer=tf_writer,
                                              save_model_frequency=save_model_frequency,
                                              saved_model_dir=saved_model_dir))
            else:
                trajectory_evaluator = \
                    DeterministicTrajectoryEvaluator(
                        reward_function=reward_function,
                        system_dynamics_handler=dynamics_handler)
        super(MPCPolicy, self).__init__(trajectory_evaluator=
                                        trajectory_evaluator)
        if optimizer is None:
            if num_agents is None:
                raise Exception("Please Specify Num Of Agents in the MPC")
            if optimizer_name == 'CEM':
                from blackbox_mpc.optimizers.cem import CEMOptimizer
                optimizer = CEMOptimizer(env_action_space=env_action_space,
                                         env_observation_space=env_observation_space,
                                         num_agents=num_agents,
                                         **optimizer_args)
            elif optimizer_name == 'CMA-ES':
                from blackbox_mpc.optimizers.cma_es import CMAESOptimizer
                optimizer = CMAESOptimizer(env_action_space=env_action_space,
                                           env_observation_space=env_observation_space,
                                           num_agents=num_agents,
                                           **optimizer_args)
            elif optimizer_name == 'PI2':
                from blackbox_mpc.optimizers.pi2 import PI2Optimizer
                optimizer = PI2Optimizer(env_action_space=env_action_space,
                                         env_observation_space=env_observation_space,
                                         num_agents=num_agents,
                                         **optimizer_args)
            elif optimizer_name == 'PSO':
                from blackbox_mpc.optimizers.pso import PSOOptimizer
                optimizer = PSOOptimizer(env_action_space=env_action_space,
                                         env_observation_space=env_observation_space,
                                         num_agents=num_agents,
                                         **optimizer_args)
            elif optimizer_name == 'SPSA':
                from blackbox_mpc.optimizers.spsa import SPSAOptimizer
                optimizer = SPSAOptimizer(env_action_space=env_action_space,
                                          env_observation_space=env_observation_space,
                                          num_agents=num_agents,
                                          **optimizer_args)
            elif optimizer_name == 'RandomSearch':
                from blackbox_mpc.optimizers.random_search import RandomSearchOptimizer
                optimizer = RandomSearchOptimizer(env_action_space=env_action_space,
                                                  env_observation_space=env_observation_space,
                                                  num_agents=num_agents,
                                                  **optimizer_args)
        self._optimizer = optimizer
        self._tf_writer = tf_writer
        self._trajectory_evaluator = trajectory_evaluator
        self._optimizer.set_trajectory_evaluator(trajectory_evaluator)
        self._act_call_counter = 0
        return

    def act(self,  observations, t, exploration_noise=False):
        """
        This is the act function for the model predictive control policy, which should be called to provide the action
        to be executed at the current time step.


        Parameters
        ---------
        observations: tf.float32
            Defines the current observations received from the environment.
        t: tf.float32
            Defines the current timestep.
        exploration_noise: bool
            Defines if exploration noise should be added to the action to be executed.


        Returns
        -------
        action: tf.float32
            The action to be executed for each of the runner (dims = runner X dim_U)
        next_observations: tf.float32
            The next observations predicted using the dynamics function learned so far.
        rewards_of_next_state: tf.float32
            The predicted reward if the action was executed using the predicted observations.
        """
        batched_observations = np.array(observations)
        if len(observations.shape) == 1:
           batched_observations = np.tile(np.expand_dims(observations, 0),
                                          (self._optimizer._num_agents, 1))
        t = tf.constant(t, dtype=tf.int32)
        current_state = tf.cast(batched_observations, dtype=tf.float32)
        mean, next_state, rewards_of_next_state = self._optimizer(current_state, t,
                                                                  tf.constant(exploration_noise,
                                                                             dtype=tf.bool))
        log_results = False
        if log_results:
            if self._tf_writer is not None:
                with self._tf_writer.as_default():
                    tf.summary.scalar('rewards/predicted_reward', tf.reduce_mean(rewards_of_next_state),
                                      step=self._act_call_counter)
        next_observations = next_state
        result_action = mean.numpy()
        next_observations = next_observations.numpy()
        self._act_call_counter += 1
        if len(observations.shape) == 1:
            result_action = result_action[0]
            next_observations = next_observations[0]
            rewards_of_next_state = rewards_of_next_state[0]
        return result_action, next_observations, rewards_of_next_state

    def reset(self):
        """
        This is the reset function for the model predictive control policy, which should be called at the beginning of
        the episode.
        """
        self._optimizer.reset()

    def switch_optimizer(self, optimizer=None, optimizer_name='',
                         **optimizer_args):
        """
        This function is used to switch the optimizer of model predictive control policy.

        Parameters
        ----------
        optimizer: OptimizerBaseClass
            Optimizer to be used that optimizes for the best action sequence and returns the first action.
        optimizer_name: str
            optimizer name between in ['CEM', 'CMA-ES', 'PI2', 'RandomSearch', 'PSO', 'SPSA'].
        optimizer_args: args
            other arguments specific to the optimizer.
        """
        if optimizer is None:
            if optimizer_name == 'CEM':
                from blackbox_mpc.optimizers.cem import CEMOptimizer
                optimizer = CEMOptimizer(env_action_space=self._optimizer._env_action_space,
                                         env_observation_space=self._optimizer._env_observation_space,
                                         num_agents=self._optimizer._num_agents,
                                         **optimizer_args)
                self._optimizer = optimizer
            elif optimizer_name == 'CMA-ES':
                from blackbox_mpc.optimizers.cma_es import CMAESOptimizer
                optimizer = CMAESOptimizer(
                    env_action_space=self._optimizer._env_action_space,
                    env_observation_space=self._optimizer._env_observation_space,
                    num_agents=self._optimizer._num_agents,
                    **optimizer_args)
                self._optimizer = optimizer
            elif optimizer_name == 'PI2':
                from blackbox_mpc.optimizers.pi2 import PI2Optimizer
                optimizer = PI2Optimizer(env_action_space=self._optimizer._env_action_space,
                                         env_observation_space=self._optimizer._env_observation_space,
                                         num_agents=self._optimizer._num_agents,
                                         **optimizer_args)
                self._optimizer = optimizer
            elif optimizer_name == 'PSO':
                from blackbox_mpc.optimizers.pso import PSOOptimizer
                optimizer = PSOOptimizer(env_action_space=self._optimizer._env_action_space,
                                         env_observation_space=self._optimizer._env_observation_space,
                                         num_agents=self._optimizer._num_agents,
                                         **optimizer_args)
                self._optimizer = optimizer
            elif optimizer_name == 'SPSA':
                from blackbox_mpc.optimizers.spsa import SPSAOptimizer
                optimizer = SPSAOptimizer(
                    env_action_space=self._optimizer._env_action_space,
                    env_observation_space=self._optimizer._env_observation_space,
                    num_agents=self._optimizer._num_agents,
                    **optimizer_args)
                self._optimizer = optimizer
            elif optimizer_name == 'RandomSearch':
                from blackbox_mpc.optimizers.random_search import \
                    RandomSearchOptimizer
                optimizer = RandomSearchOptimizer(
                    env_action_space=self._optimizer._env_action_space,
                    env_observation_space=self._optimizer._env_observation_space,
                    num_agents=self._optimizer._num_agents,
                    **optimizer_args)
                self._optimizer = optimizer
        else:
            self._optimizer = optimizer
        self._optimizer.set_trajectory_evaluator(self._trajectory_evaluator)
        return



import tensorflow as tf


@tf.function
def default_transform_targets(current_state, next_state):
    """
    This is the default transform targets function used, which preprocesses the targets of the network before training the
    dynamics function using the inputs and targets. The default one is (target = next_state - current_state).

    Parameters
    ---------
    current_state: tf.float32
        The current_state has a shape of (Batch X dim_S)
    next_state: tf.float32
        The next_state has a shape of (Batch X dim_S)
    """
    return next_state - current_state


@tf.function
def default_inverse_transform_targets(current_state, delta):
    """
    This is the default inverse transform targets function used, which reverses the preprocessing of  the targets of
    the dynamics function to obtain the real current_state not the relative one,
    The default one is (current_state = target + current_state).

    Parameters
    ---------
    current_state: tf.float32
        The current_state has a shape of (Batch X dim_S)
    delta: tf.float32
        The delta has a shape of (Batch X dim_S) which is equivilant to the target of the network.
    """
    return delta + current_state



from blackbox_mpc.policies.random_policy import RandomPolicy
from blackbox_mpc.policies.mpc_policy import MPCPolicy
from blackbox_mpc.policies.model_based_base_policy import ModelBasedBasePolicy
from blackbox_mpc.policies.model_free_base_policy import ModelFreeBasePolicy